/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

//Code for dealing with command-line options
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <math.h>
#include <time.h>
#include <stdarg.h>
using namespace std;

#include "ErrorHandler.h"
#include "Macros.h"
#include "Miscellaneous.h"
#include "filesearch.cpp"

#define CONCATENATEDIRS(A,B) A/B
#define VARIABLEFILE MAKESTRING(CONCATENATEDIRS(ROOTDIR,DEFAULTVARIABLES))

//Functions to use as function pointers

void noprnt(ostream &o,void *address){
};

void prntstr(ostream &o,void *address){
  o<<*(const char **)address;
};

void prntint(ostream &o,void *address){
  o<<*(int *)address;
};

void prntld(ostream &o,void *address){
  o<<*(long double*)address;
};

void prntdbl(ostream &o,void *address){
  o<<*(double*)address;
};

void prntflt(ostream &o,void *address){
  o<<*(float*)address;
};

int setflag(const char *val,void *address){
  return 0;
};

int setstr(const char *val,void *address){
  if(val==NULL){
    *(const char **)address=NULL;
  }else{
    //    cout<<val<<"\n";
    char *cpy=new char[strlen(val)+1];
    strcpy(cpy,val);
    *(const char **)address=cpy;
    //    cout<<*(const char **)address<<"\n";
  };
  return 1;
};

void prntpth(ostream &o,void *address){
  for(int n=0;*(*(char ***)address+n);n++){
    o<<*(*(char ***)address+n)<<"\n";
  };
};

int setint(const char *val,void *address){//Should allow macros.
  *(int*)address=atoi(val);
  return 1;
};

int setflt(const char *val,void *address){
  *(float*)address=atof(val);
  return 1;
};

int setdbl(const char *val,void *address){
  *(double*)address=atof(val);
  return 1;
};

int setld(const char *val,void *address){
  *(long double*)address=atof(val);
  return 1;
};

int brkstr(const char *val, void *address){
  //Reads something like a path variable, and converts it to a list of
  //strings, by separating on the character : (if not escaped)
  int n=1;
  for(int i=0;*(val+i);i++){
    if(*(val+i)==':'){
      n++;
    };
    if(*(val+i)=='\\'){
      i++;
      if(!*(val+i)){//Can't escape null character.
	break;
      };
    };
  };
  *(char ***)address=new char *[n+1];
  const char *v=val;
  for(int j=0;j<n;j++){
    int i=0;
    for(;*(v+i)&&*(v+i)!=':';i++){
      if(*(v+i)=='\\'){
	i++;
	if(!*(v+i)){//Can't escape null character.
	  break;
	};
      };
    };
    *(*(char ***)address+j)=new char[i+1];
    int kk=0;
    for(int k=0;k<i;k++){
      if(*(v+k)=='\\'){
	k++;
      };								\
      *(*(*(char ***)address+j)+kk)=*(v+k);
      kk++;
    };
    *(*(*(char ***)address+j)+kk)='\0';
    v+=i+1;
  };
  *(*(char ***)address+n)=NULL;
  //  prntpth(cout,address);
  return 1;
};

void delpth(void *address){
  for(int n=0;*(*(char ***)address+n);n++){
    delete[] *(*(char ***)address+n);
  };
  delete[] *(char ***)address;
  delete[] (char ***)address;
};

void delstr(void *address){
  delete[] *(char **)address;
  delete[] (char **)address;
};

void delint(void *address){
  delete[] (int *)address;
};

void delld(void *address){
  delete[] (long double *)address;
};

void delflt(void *address){
  delete[] (float *)address;
};

void deldbl(void *address){
  delete[] (double *)address;
};

void *makevar(char *type);

// ---------------------------------------------------------------------------

class variable{//Used for storing system variables
public:
  char *name;
  char *tag;//commandline option tag
  void *value;
  int (*setv)(const char *val,void *address);
  void (*delv)(void *address);
  void (*prnt)(ostream& o,void *address);
  int setval(const char *val){if(value==NULL){return 0;}else{return (*setv)(val,value);};};
  variable(const char *n, const char *t,void* address,int (*set)(const char *val,void *address),void (*print)(ostream& o,void *address),void (*del)(void *add)){name=new char[strlen(n)+1];tag=new char[strlen(t)+1];strcpy(name,n);strcpy(tag,t);value=address;setv=set;prnt=print;delv=del;};
  ~variable(){if(delv!=NULL){delv(value);};delete[] name; delete[] tag;};
};

class varlist{//linked list of variables.
  variable *x;
  varlist *next;
public:
  varlist(){x=NULL;next=NULL;};
  void *seekval(const char *vname){if(!strcmp(vname,x->name)){return x->value;}else if(next==NULL){return NULL;}else{return next->seekval(vname);};};
  variable *seektag(char *tg){if(!strcmp(tg,x->tag)){return x;}else if(next==NULL){return NULL;}else{return next->seektag(tg);};};
  void setvals(int argc,char *argv[]);
  void addvar(const char *name,const char *tag,void *address,int (*setv)(const char *val,void *address),const char *init,void (*prnt)(ostream& o,void *address),void (*delv)(void *add)=NULL);
  void remove(){if(next!=NULL){next->remove();delete next;};if(x!=NULL){delete x;};};
  //  ~varlist(){this->remove();};
  void addvars(ifstream& in);
  void setupvars();
  friend ostream &operator <<(ostream &o,varlist &v);
};

varlist vars;

void removevarlist(){
  vars.remove();
};

ostream &operator <<(ostream &o,variable& v){
  o<<v.name<<"=\"";
  if(v.value==NULL){
    o<<"undefined";
  }else{
    (*v.prnt)(o,v.value);
  };
  o<<"\"\n";
  return o;
};


ostream &operator <<(ostream &o,varlist &v){
  if(v.x!=NULL){
    o<<*v.x;
  };
  if(v.next!=NULL){
    o<<*v.next;
  };
  return o;
};

class fnptrs{
public:
  int (*setv)(const char *,void *);
  void (*prnt)(ostream&,void*);
  void (*delv)(void *);
};

fnptrs getfuns(const char *type);

void varlist::addvars(ifstream& in){
  char *tag;
  char *type;
  char *init;
  fnptrs f;
  void *address;
  while(in.good()){
    char *line=readstring(in,"\n",'\\');
    //Format: variable name, [variable tag], variable type, initial value
    //Should add escape character.
    int i;
    for(i=0;*(line+i)&&*(line+i)!=' ';i++);
    for(;*(line+i)==' ';i++){
      *(line+i)='\0';
    };
    if(!*(line+i)){//Just a name, make it a string variable with tag equal to name plus -
      tag=new char[strlen(line)+2];
      *tag='-';
      for(unsigned int j=0;j<strlen(line)+1;j++){
	*(tag+j+1)=*(line+j);
      };
      this->addvar(line,tag,new char[20],setstr,"",noprnt,delstr);
    }else{
      int j=i;
      if(*(line+j)=='-'){//tag
	tag=line+j+1;
	for(;*(line+j)&&*(line+j)!=' ';j++);
	for(;*(line+j)==' ';j++){
	  *(line+j)='\0';
	};
	if(!*(line+j)){//Just name and tag
	  this->addvar(line,line+i,new char[20],setstr,"",noprnt,delstr);
	  continue;
	};
      }else{
	tag=new char[strlen(line)+2];
	*tag='-';
	for(unsigned int j=0;j<strlen(line)+1;j++){
	  *(tag+j+1)=*(line+j);
	};
      };
      type=line+j;
      for(;*(line+j)&&*(line+j)!=' ';j++);
      for(;*(line+j)==' ';j++){
	*(line+j)='\0';
      };
      init=line+j;
      f=getfuns(type);
      address=makevar(type);
      this->addvar(line,tag,address,f.setv,init,f.prnt,f.delv);
    };
    for(char c='\n';(char) in.peek()=='\n';in.get(c));
    delete[] line;
  };
};

void varlist::addvar(const char *name,const char *tag,void *address,int (*setv)(const char *val,void *address),const char *init,void (*prnt)(ostream& o,void *address),void (*delv)(void *add)){
  //add the variable in second place, for simplicity.
  if(x==NULL){
    x=new variable(name,tag,address,setv,prnt,delv);    
    x->setval(init);
  }else{
    varlist *tmp=new varlist;
    tmp->x=new variable(name,tag,address,setv,prnt,delv);    
    tmp->x->setval(init);
    tmp->next=next;
    next=tmp;
  };
};

void varlist::setupvars(){
  info("Setting up default variables.",msgcode(1,0));
  ifstream vbls(VARIABLEFILE);
  this->addvars(vbls);
  vbls.close();
};

int setifstream(const char *val,void *address);

void varlist::setvals(int argc,char *argv[]){
  //Add default variables file ".variables"
  strlist flags;
  ifstream fl;
  setifstream(".flags",&fl);
  //Should change this to allow user to specify a filename. Actually,
  //should compile this into the program at compile time.
  if(fl.is_open()){
    while(fl.good()){
      char *ff=readnonemptystring(fl," \t\r\n");
      flags.insert(ff);
      delete[] ff;
    };
  };
  char *tg;
  const char *val;
  for(int i=1;i<argc;i++){
    variable *v;
    if(*(argv[i])=='-'){//option file
      if(*(argv[i]+1)=='-'){//long option name terminated by space
	int l=2;
	for(;*(argv[i]+l);l++);
	tg=new char[l];
	strcpy(tg,(argv[i]+1));
	val=argv[++i];
      }else{//single character option name
	tg=new char[2];
	*tg=*(argv[i]+1);
	*(tg+1)='\0';
	if(*(argv[i]+2)){//value included in this argument
	  val=(argv[i]+2);
	}else{
	  val=(argv[++i]);
	}
      };
    }else{
      tg=new char[1];
      *tg='\0';
      val=argv[i];
    };
    v=this->seektag(tg);
    if(v==NULL){
      if(!strcmp(tg,"-variables")){//Read more command line options and
				  //variables from a file.
	ifstream vbls(val);
	this->addvars(vbls);
	vbls.close();
      }else if(flags.search(tg)!=-1){
	this->addvar("",tg,NULL,setflag,val,NULL,NULL);	
	i--;
      }else{
	this->addvar("",tg,new const char *[1],setstr,val,prntstr,delstr);
      };
    }else{
      if(!v->setval(val)){//no value for tag --- just on-off
	i--;
      };
    };
    delete[] tg;
  };
  flags.remove();
};

int stlen(const char **pt){
  int l=0;
  for(int i=0;*(pt+i);i++){
    l+=strlen(*(pt+i))+2;
  };
  return l;
};

void pathtostr(const char **pt, char *out){
  for(int i=0;*(pt+i);i++){
    strcat(out,*(pt+i));
    strcat(out,"\n");
  };
};

int setifstream(const char *val,void *address){
  ifstream *str=(ifstream *)address;
  void *pt=vars.seekval("path");
  if(pt==NULL){
    ifstream test(val);
    if(!test.good()){
      //      cout<<"Could not open file: \""<<val<<"\"\n";
      char *x=new char[30+strlen(val)];
      strcpy(x,"Could not open file \"");
      strcat(x,val);
      strcat(x,"\"");
      recoverableError(x);//Should add a new file menu.
    }else{
      test.close();
    };
    if(str->is_open()){
      str->close();
    };
    str->open(val);
    if(!str->good()){
      char *x=new char[30+strlen(val)];
      strcpy(x,"IO error openning file \"");
      strcat(x,val);
      strcat(x,"\"");
      recoverableError(x);//Should add a new file menu.      
    };
  }else{
    if(!search(*(const char ***)pt,val,*str)||!str->good()){
      /*      if(str->bad()){
	cout<<"Stream is bad.\n";
      };
      if(str->eof()){
	cout<<"End of file.\n";
      };
      if(str->fail()){
	cout<<"Stream operation failed.\n";
	};*/
      char *x=new char[90+strlen(MAKESTRING(ROOTDIR))+strlen(val)+stlen(* (const char ***)pt)];//
      strcpy(x,"Could not open file: \"");
      strcat(x,val);
      strcat(x,"\"\n\n");
      strcat(x,"search path=\n");
      pathtostr(*(const char ***)pt,x);
      strcat(x,"root seach directory=");
      strcat(x,MAKESTRING(ROOTDIR));
      recoverableError(x);//Should add a new file menu.      
    };
  };
  return 1;
};

int *match(int i,int j,int *graph){
  //finds a matching in the bipartite graph if one exists.  That is,
  //for every element less than i, it attempts to find a different
  //element less than j, which is joined to it.

  //primitive algorithm, might need to improve on it later.

  for(int k=0;k<i;k++){
    int l=0;
    for(;l<j&&*(graph+k*j+l)!=1;l++);
    if(l==j){//need to backtrack.
      if(k==0){
	return NULL;
      };
      int jj=0;
      for(int ii=0;ii<i;ii++){
	if(*(graph+k*j+ii)==-1){
	  *(graph+k*j+ii)=1;
	};
      };
      k--;
      for(;*(graph+k*j+jj)!=k;jj++);
      for(int kk=0;kk<i;kk++){
	if(*(graph+kk*j+jj)==k){
	  *(graph+kk*j+jj)=1;
	};
      };
      *(graph+k*j+jj)=-1;
      k--;
    }else{//try this matching;
      for(int kk=0;kk<i;kk++){
	if(*(graph+kk*j+l)==1){
	  *(graph+kk*j+l)=k;
	};
      };
    };
  };
  int *ans=new int[i];
  for(int ii=0;ii<i;ii++){
    int jj=0;
    for(;*(graph+ii*j+jj)!=ii;jj++);
    *(ans+ii)=jj;
  };
  return ans;
};

int chooseifstreams(char *basename,int n,...){
  //n is the number of streams. The remaining arguments begin with lists of extensions
  va_list argptr;
  va_start(argptr,n);
  void *pt=vars.seekval("path");
  const char *const*path;
  if(pt==NULL){
    char **pth=new char*[1];
    *pth=new char[1];
    **pth='\0';
    path=pth;
  }else{
    path=*(const char ***)pt;
  };
  char **found=new char*[n];
  for(int i=0;i<n;i++){
    const char *const*ex=va_arg(argptr,const char*const*);
    *(found+i)=searchall(path,basename,ex);
  };
  int num=0;
  for(int i=0;i<n;i++){
    for(char *fnd=*(found+i);strlen(fnd);fnd+=strlen(fnd)+1){
      int j=0;
      for(;j<i;j++){
	char *fnd2=*(found+j);
	for(;strlen(fnd2)&&strcmp(fnd,fnd2);fnd2+=strlen(fnd2)+1);
	if(strlen(fnd2)){
	  break;
	};
      };
      if(j==i){
	num++;
      };
    };
  };
  char **allfnd=new char*[num];
  int *graph=new int[num*n];
  for(int x=0;x<num*n;x++){
    *(graph+x)=0;
  };
  int nm=0;
  for(int i=0;i<n;i++){
    for(char *fnd=*(found+i);strlen(fnd);fnd+=strlen(fnd)+1){
      int j=0;
      for(;j<i;j++){
	char *fnd2=*(found+j);
	int ps=0;
	for(;strlen(fnd2)&&strcmp(fnd,fnd2);fnd2+=strlen(fnd2)+1){
	  ps++;
	};
	if(strlen(fnd2)){
	  int ii=0;
	  for(int jj=0;jj<=ps;jj++){
	    for(;*(graph+j*num+ii)==0;ii++);
	    ii++;
	  };
	  *(graph+i*num+ii-1)=1;
	  break;
	};
      };
      if(j==i){
	*(allfnd+nm)=new char[strlen(fnd)+2];
	strcpy(*(allfnd+nm),fnd);
	*(graph+i*num+nm)=1;
	nm++;
      };
    };
  };
  /*  cout<<"Possible files:\n";
  for(int i=0;i<num;i++){
    cout<<*(allfnd+i)<<"\n";
  };
  for(int i=0;i<n;i++){
    for(int j=0;j<num;j++){
      cout<<*(graph+i*num+j)<<" ";
    };
    cout<<"\n";
    };*/
  int *mt=match(n,num,graph);
  if(mt==NULL){
    va_end(argptr);
    return 0;
  };
  /*  for(int i=0;i<n;i++){
    cout<<*(mt+i)<<" ";
  };
  cout<<"\n";*/
  for(int i=0;i<n;i++){
    ifstream *in=va_arg(argptr,ifstream*);
    //    cout<<"Openning file \""<<*(allfnd+*(mt+i))<<"\"\n";
    if(in->is_open()){
      in->close();
    };
    in->open(*(allfnd+*(mt+i)));
    if(!in->is_open()||!in->good()){
      char *x=new char[30+strlen(*(allfnd+*(mt+i)))];
      strcpy(x,"IO error openning file \"");
      strcat(x,*(allfnd+*(mt+i)));
      strcat(x,"\"");
      recoverableError(x);//Should add a new file menu.      
    };
  };
  for(int i=0;i<n;i++){
    delete[] *(found+i);
  };
  delete[] found;
  delete[] mt;
  delete[] allfnd;
  delete[] graph;
  if(pt==NULL){
    delete[] path;
  };
  va_end(argptr);
  return 1;
};

fnptrs getfuns(const char *type){
  fnptrs f;
  if(!strcmp(type,"string")){
    f.setv=setstr;
    f.prnt=prntstr;
    f.delv=delstr;
  }else if(!strcmp(type,"int")){
    f.setv=setint;
    f.prnt=prntint;
    f.delv=delint;
  }else if(!strcmp(type,"float")){
    f.setv=setflt;
    f.prnt=prntflt;
    f.delv=delflt;
  }else if(!strcmp(type,"dbl")){
    f.setv=setdbl;
    f.prnt=prntdbl;
    f.delv=deldbl;
  }else if(!strcmp(type,"ld")){
    f.setv=setld;
    f.prnt=prntld;
    f.delv=delld;
  }else if(!strcmp(type,"path")){
    f.setv=brkstr;
    f.prnt=prntpth;
    f.delv=delpth;
  }else if(!strcmp(type,"flag")){
    f.setv=setflag;
    f.prnt=NULL;
    f.delv=NULL;
  }else{
    f.setv=NULL;
    f.prnt=NULL;
    f.delv=NULL;
  };
  return f;
};

void *makevar(char *type){
  if(!strcmp(type,"string")){
    return new char* [1];
  }else if(!strcmp(type,"int")){
    return new int [1];
  }else if(!strcmp(type,"float")){
    return new float [1];
  }else if(!strcmp(type,"dbl")){
    return new double [1];
  }else if(!strcmp(type,"ld")){
    return new long double [1];
  }else if(!strcmp(type,"path")){
    return new char**[1];
  }else{
    return NULL;
  };
};

