/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <iostream>
#include <stdlib.h>
#include <stdarg.h>
using namespace std;
#include "ErrorHandler.h"
#include "Matrix.h"
#include "Parameters.h"
#include "Sequence.h"
#include "Tree.h"
#include "Miscellaneous.h"
#include "CommandLine.h"
#include "MixTreeLikelihood.h"

extern varlist vars;

int match(const char *a,const char *str){
  int m=0;
  for(;*(str+m)&&*(a+m)&&(*(a+m)==*(str+m)||*(a+m)==toupper(*(str+m)));m++);
  return m;
};

void params::usematrix(const Realmatrix &Q){
  numpars=0;
  if(XX!=NULL){
    delete[] XX;
    XX=NULL;
  };
  if(mask!=NULL){
    delete mask;
    mask=NULL;
  };
  A=Q;//Need to symmetrise and get pi.
  setupp=1;
};

void params::getmask(int n,ifstream& in){
  for(int i=1;i<n;i++){
    dummyreadmatrix(in);
  };
  for(;in.good()&&!(in.eof())&&in.peek()=='\n';in.get());
  if(mask!=NULL){
    delete mask;
  };
  mask=new Realmatrix(in); 
};

void params::setup(ifstream& in){
  //Should change the function for matrices to allow a matrix to be
  //input from a stream. When that is done, can change this function to
  //accept any istream.
  //Reads in the fixed matrices from the input stream.
  //  numpars=NUMPARS;

  streampos st=in.tellg();
  int maskpos=-1;//No mask
  int nm1=0;
  int nm2=0;
  for(char c='\n';c=='\n'&&in.good()&&!(in.eof());in.get(c));
  in.seekg(-1,ios::cur);
  char *firstline=readstring(in,"\n");
  //Format of first line:
  //nummats [mask] nummats
  char *pos=firstline;
  for(;*pos==' '||*pos=='\t';pos++);//strip leading spaces
  int msk=match(pos,"mask");
  pos+=msk;
  for(;*pos==' '||*pos=='\t';pos++);//strip leading spaces
  if(msk>0){
    maskpos=0;
    nm2=atoi(pos);
  }else{
    nm1=atoi(pos);
    for(;isdigit(*pos);pos++);
    for(;*pos==' '||*pos=='\t';pos++);//strip leading spaces
    msk=match(pos,"mask");
    if(msk>0){
      maskpos=nm1;
      pos+=msk;
      for(;*pos==' '||*pos=='\t';pos++);//strip leading spaces
      nm2=atoi(pos);
      for(;isdigit(*pos);pos++);
      for(;*pos==' '||*pos=='\t';pos++);//strip leading spaces
    };
    if(isdigit(*pos)||*pos=='-'){//first line isn't correctly
				 //formatted. probably first line of a
				 //matrix.
      maskpos=-2;
    };
  };
  if(maskpos==-2){/* Matrices start on first line */
    in.seekg(st);
    if(in.fail()){//Can't return to chosen position.
      char* msg=new char[50+strlen(Filename)];
      sprintf(msg,"IO error reading parameter file \"%s\".",Filename);
      recoverableError(msg);
      //Could improve this later in order to allow recovery.
    };
    nm1=-1;
  };
  delete[] firstline;

  variable* rec=vars.seektag("-numpars");
  if(rec!=NULL){
    numpars=atoi(*(char**)(rec->value));
  };

  if(nm1==-1){
    nm1=numpars;
    nm2=0;
  };

  if(nm1+nm2<numpars){//Some problem with format of file
    if(rec!=NULL){
      char msg[100];
      sprintf(msg,"Too few matrices in parameter file.\n Reducing number of parameters in the model to %i.",nm1+nm2);
      warning(msg);
    };
    numpars=nm1+nm2;
  }else if(nm1+nm2>numpars){
    if(rec!=NULL){
      nm2=numpars-nm1;
    }else{
      numpars=nm2+nm1;
    };
  };

  rec=vars.seektag("-parameterselection");
  if(rec!=NULL){
    int *sel=selection(*(char **)rec->value);
    sortlist(sel);
    for(numpars=0;*(sel+numpars)>0;numpars++){
      (*(sel+numpars))--;
    };
    numpars++;
    XX=new Realmatrix[numpars];
    XX->sz=NumCodons;
    XX->entries=new long double[NumCodonsSq];
    for(int i=0;i<NumCodonsSq;i++){
      *(XX->entries+i)=1;
    };
    for(int i=0;i<NumCodons;i++){
      *(XX->entries+i*(NumCodons+1))=1-NumCodons;
    };
    int p=0;
    for(int i=0;*(sel+p)>=0&&in.good();i++){
      if(i==maskpos){
	if(!vars.seektag("-nomask")){
	  mask=new Realmatrix(in);
	}else{
	  dummyreadmatrix(in);
	};
      };
      if(*(sel+p)==i){
	*(XX+p+1)=RealmatrixT(in);
	p++;
      }else{
	dummyreadmatrix(in);
      };
    };
    if(!in.good()){
      char message[100];
      sprintf(message,"Insufficient parameters in file. Stopped after reading parameter number %i.",*(sel+p));
      warning(message);
      numpars=p;
    };
    delete[] sel;
  }else{
    for(char c='\n';c=='\n'&&in.good()&&!(in.eof());in.get(c));
    in.seekg(-1,ios::cur);
    numpars++;
    XX=new Realmatrix[numpars];//Should add a way to specify no. of
    XX->sz=NumCodons;
    XX->entries=new long double[NumCodonsSq];
    for(int i=0;i<NumCodonsSq;i++){
      *(XX->entries+i)=1;
    };
    for(int i=0;i<NumCodons;i++){
      *(XX->entries+i*(NumCodons+1))=1-NumCodons;
    };
    //params in the file.
    for(int i=0;i<nm1;i++){
      if(i<numpars-1){
	*(XX+i+1)=RealmatrixT(in);
      }else{
	dummyreadmatrix(in);
      };
      for(char c='\n';c=='\n'&&in.good()&&!(in.eof());in.get(c));
      in.seekg(-1,ios::cur);
      if(in.good()){
	in.seekg(-1,ios::cur);
      };
    };
    if(maskpos>=0){
      if(!(vars.seektag("-nomask"))){
	mask=new Realmatrix(in);
      }else{
	dummyreadmatrix(in);
      };
      for(char c='\n';c=='\n'&&in.good()&&!(in.eof());in.get(c));
      for(int i=0;i<nm2;i++){
	*(XX+i+nm1+1)=RealmatrixT(in);
	for(char c='\n';c=='\n'&&in.good()&&!(in.eof());in.get(c));
	in.seekg(-1,ios::cur);
	if(in.good()){
	  in.seekg(-1,ios::cur);
	};
      };
    };
  };
  //  A=Realmatrix(in);
  A=normalisemats(XX,numpars);//Need to sort out
				     //orthonormalisation for Mixture
				     //models.
  atexit(unsetupParams);
  setupp=1;
};

void params::readpars(istream &in){//Reads in a starting set of parameters.
  /*  for(int i=0;i<NumCodons&&in.good();i++){
    char *a=readnonemptystring(in," \n\t\r");
    *(pi+i)=atof(a);    
    delete[] a;
    };

pi normally found empirically.
  */
  for(int i=0;i<NumCodons;i++){
    *(pi+i)=1.0L/61;
  };
  long double *cf=new long double[numpars];
  for(int i=0;i<numpars;i++){
    char *a=readnonemptystring(in," \n\t\r");
    *(cf+i)=atof(a);    
    delete[] a;    
  };
  ((Realmatrix)A.inverse()).act(cf,coeff);
  delete[] cf;
  this->setR();
  long double *d=new long double[NumCodons];
  C=R.diagonalisesymmetric(d);
  delete[] d;
};

void sequence::assign(ifstream &in){
  char *a=new char[25];
  *a=' ';
  for(;*a==' '||*a=='\t';in.get(*a)); //strip leading spaces.
  in.get(a+1,24,' ');
  in.clear();
  length=atoi(a);
  seq=new char[length+2];
  *seq=' ';
  for(;in.good()&&*seq==' ';in.get(*seq)); //strip leading spaces.
  in.get(seq+1,length);
  if(strlen(seq)<(unsigned)length){//Short sequence
    cerr<<"Short sequence. Aborting.\n";
    exit(1);
  };
  length/=CodonLength;
  //continue to newline;
  for(char t=' ';t!='\n'&&in.good();in.get(t));
  delete[] a;
};

void sequence::assign(ifstream &in,int len){
  char *a=new char[25];
  *a=' ';
  for(;*a==' '||*a=='\t';in.get(*a)); //strip leading spaces.
  if(isdigit(*a)){
    in.get(a+1,24,' ');
    length=atoi(a);
    seq=new char[length+2];
    *seq=' ';
  }else{
    length=len;
    seq=new char[length+2];
    *seq=*a;
  };
  for(;*seq==' ';in.get(*seq)); //strip leading spaces.
  for(int i=1;i<length;i++){
    in.get(*(seq+i));
    if(*(seq+i)==' '||*(seq+i)=='\t'||*(seq+i)=='\n'||*(seq+i)=='\r'){
      i--;
    };
    if(!in.good()){
      throw 1;
    };
  };
  *(seq+length)='\0';
  if(strlen(seq)<(unsigned)length){//Short sequence
    cerr<<"Short sequence. Aborting.\n";
    exit(1);
  };
  length/=CodonLength;
  //continue to newline;
  for(char t=' ';t!='\n'&&in.good();in.get(t));
  delete[] a;
};

tree::tree(int l,const char *s){
  char *rubbish;
  up=new long double[NumCodons];
  up2=new long double[NumCodons];
  up3=new long double[NumCodons];
  down=new long double[NumCodons];
  down2=new long double[NumCodons];
  down3=new long double[NumCodons];
  n=0;
  int depth=0;
  for(int i=0;i<l;i++){
    if(*(s+i)=='('){
      depth++;
    }else if(*(s+i)==')'){
      depth--;
    }else if(*(s+i)==','&&depth==1){
      n++;
    };
  };
  for(;l>0&&(*s==' '||*s=='\t'||*s=='\r'||*s=='\n');s++){
    l--;
  };
  if(n==0){
    int len=l+1;
    if(*s=='('){
      len-=2;
      s++;
    };
    label=new char[len];
    for(int i=0;i<len-1;i++){
      *(label+i)=*(s+i);
    };
    *(label+len-1)='\0';
    return;
  }else{
    label=NULL;
    n++;
  };
  if(depth!=0){
    char dep[5];//If parantheses are mismatched worse than this, we
		//have a problem.
    sprintf(dep,"%d",depth);
    char *x=new char[100+strlen(s)];
    strcpy(x,"Mismatched parantheses in tree definition. \"");
    strcat(x,s);
    strcat(x,"\"\n");
    strcat(x,"a total of ");
    strcat(x,dep);
    strcat(x,"parantheses are unmatched.");
    recoverableError(x);
    delete[] x;
  };
  depth=0;
  length=new long double[n];
  subtree=new tree[n];
  int j=0;
  int kold=1;
  for(int i=0;i<l;i++){
    for(;*(s+i)==' '||*(s+i)=='\t'||*(s+i)=='\r'||*(s+i)=='\n';i++);
    if(*(s+i)=='('){
      depth++;
    }else if(*(s+i)==')'){
      depth--;
    }else if(*(s+i)==','&&depth==1){
      int k=i;
      for(;k>=kold&&(*(s+k)!=':');k--);
      if(*(s+k)==':'){
	*(length+j)=strtod(s+k+1,&rubbish);
      }else{
	*(length+j)=1;
	k=i;
      };
      *(subtree+j)=tree(k-kold,s+kold);
      j++;
      kold=i+1;
    };
  };
  int k=l-1;
  for(;k>=0&&*(s+k)!=':';k--);
  if(*(s+k)==':'){
    *(length+j)=strtod(s+k+1,&rubbish);
  }else{
    *(length+j)=1;
    k=l-1;
  };
  *(subtree+j)=tree(k-kold,s+kold);
};

tree::tree(const char *s){//Should make this more robust.
  char *rubbish;
  int l=strlen(s);
  up=NULL;//new long double[NumCodons];
  up2=NULL;//new long double[NumCodons];
  up3=NULL;//new long double[NumCodons];
  down=new long double[NumCodons];
  down2=new long double[NumCodons];
  down3=new long double[NumCodons];//Could perhaps modify this to put similar types of lists together in memory.
  n=0;
  int depth=0;
  for(int i=0;i<l;i++){
    if(*(s+i)=='('){
      depth++;
    }else if(*(s+i)==')'){
      depth--;
    }else if(*(s+i)==','&&depth==1){
      n++;
    };
  };
  for(;*s==' '||*s=='\t'||*s=='\r'||*s=='\n';s++){
    l--;
  };
  if(n==0){
    int len=l+1;
    if(*s=='('){
      len-=2;
      s++;
    };
    label=new char[len];
    for(int i=0;i<len-1;i++){
      *(label+i)=*(s+i);
    };
    *(label+len-1)='\0';
    return;
  };
  label=NULL;
  n++;
  if(depth!=0){
    char dep[5];//If parantheses are mismatched worse than this, we
		//have a problem.
    sprintf(dep,"%d",depth);
    char *x=new char[100+strlen(s)];
    strcpy(x,"Mismatched parantheses in tree definition. \"");
    strcat(x,s);
    strcat(x,"\"\n");
    strcat(x,"a total of ");
    strcat(x,dep);
    strcat(x," parantheses are unmatched.");
    recoverableError(x);
    delete[] x;
  };
  //  depth=0;
  length=new long double[n];
  subtree=new tree[n];
  int j=0;
  int kold=1;
  for(int i=0;i<l;i++){
    for(;*(s+i)==' '||*(s+i)=='\t'||*(s+i)=='\r'||*(s+i)=='\n';i++);
    if(*(s+i)=='('){
      depth++;
    }else if(*(s+i)==')'){
      depth--;
    }else if(*(s+i)==','&&depth==1){
      int k=i;
      for(;k>kold&&(*(s+k)!=':')&&(*(s+k)!=')');k--);
      if(k==kold){
	*(subtree+j)=tree(i-kold,s+kold);
	*(length+j)=1;
      }else if(*(s+k)==':'){
	*(subtree+j)=tree(k-kold,s+kold);
	*(length+j)=strtod(s+k+1,&rubbish);
      }else{
	*(subtree+j)=tree(k-kold+1,s+kold);
	*(length+j)=1;
      };
      j++;
      kold=i+1;
    };
  };
  int k=l-1;
  for(;k>kold&&*(s+k)!=':';k--);
  if(k==kold){
    *(length+j)=1;
    k=l-1;
    for(;*(s+k)!=')';k--);
    //    k--;
  }else{
    *(length+j)=strtod(s+k+1,&rubbish);
  };
  *(subtree+j)=tree(k-kold,s+kold);
  this->setextra(this->edges());
};

sequence *getdata(ifstream& f,const tree& T){
  streampos st=f.tellg();
  int numseq;
  int len;
  sequence *seq=new sequence[T.leaves()];
  char *rd=new char[50];
  int i;
  int* ass=new int[T.leaves()];
  for(int i=0;i<T.leaves();i++){
    *(ass+i)=1;
  };
  int assno=0;
  char *firstline=readstring(f,"\n");
  try{
    readline(firstline,"uU",&numseq,-1,&len);
  }catch(char x){//firstline doesn't give general information
    f.seekg(st);
    len=-1;
    numseq=-1;
  };
  //  f.clear();
  while(f.good()){
    f.get(*rd);
    for(;f.good()&&(*rd==' '||*rd=='\n'||*rd=='\r'||*rd=='\t');f.get(*rd));
    f.get(rd+1,49,' ');
    i=T.findlabel(rd);
    if(i==-1){
      for(char c=' ';c!='\n'&&f.good();f.get(c));
    }else{
      try{
	if(len==-1){
	  (seq+i)->assign(f);
	}else{
	  (seq+i)->assign(f,len);
	};
	if(*(ass+i)){       
	  *(ass+i)=0;
	  assno++;
	};
      }catch(int i){//IO error reading sequence
	char *x=new char[50+strlen(rd)];
	strcpy(x,"Problem reading sequence for species \"");
	strcat(x,rd);
	strcat(x,"\"");
	recoverableError(x);
	delete[] x;
      };
    };
  };
  if(assno<T.leaves()){//No data for some nodes
    //Should rewrite this to use error functions.
    //But need to allow T.printspecies to output a string.
    cerr<<"Unable to find all data sequences.\n";
    cerr<<"Data missing for species:\n";
    T.printspecies(cerr,ass);
    exit(1);
  };
  delete[] rd;
  delete[] ass;
  delete[] firstline;
  return seq;
};



Mixture::Mixture(const char *s,int n){
  num=n;
  this->assign(s);
};

Mixture::Mixture(istream& s,int n){
  num=n;
  this->assign(s);
};

void Mixture::assign(const char *s){
  const char *t=s;
  int rn=0;
  int *cl=new int[(num+1)*params::numpars];
  for(int i=0;i<(num+1)*params::numpars;i++){
    *(cl+i)=-1;
  };
  int nextcl=0;
  int joinrow=-1;
  int lastjoinrow=-1;
  int from=-1;
  linkedlist<int> cur;
  linkedlist<int> row;
  for(;*t;){
    for(;*t==' '||*t=='\t'||*t=='\r'||*t=='\n'||*t==';';t++);
    if(isdigit(*t)){
      rn=atoi(t);
      if(rn>=params::numpars){
	if(from==-1){
	  for(;*t&&*t!='\n'&&*t!=';';t++);
	  continue;
	}else{
	  rn=params::numpars;
	};
	//Could give surprising behaviour. Should improve this.
      };
      for(;isdigit(*t);t++);
      for(;*t==' '||*t=='\t'||*t=='\r';t++);
      lastjoinrow=joinrow;
      if(from!=-1){//Fill in from previous pattern
	if(from<rn){
	  for(int i=from+1;i<rn;i++){
	    if(*(cl+i*(num+1))==-1){
	      *(cl+i*(num+1))=1;
	      if(joinrow==-1){//New blocks for each row
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
		  };
		  if(rw->sz>0){
		    nextcl++;
		  };
		};
	      }else{//Use same blocks for all rows
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=cno;
		  };
		};
	      };
	    };
	  };
	}else{
	  for(int i=from-1;i>rn;i--){
	    if(*(cl+i*(num+1))==-1){
	      *(cl+i*(num+1))=1;
	      if(joinrow==-1){//New blocks for each row
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
		  };
		  if(rw->sz>0){
		    nextcl++;
		  };
		};
	      }else{//Use same blocks for all rows
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=cno;
		  };
		};
	      };
	    };
	  };
	};
      };
      from=-1;
    }else if(*t=='.'){
      from=rn;
      t++;
      //      for(;*t&&*t!='\n'&&*t!=';';t++);
      //      continue;
    }else if(*t=='|'){
      if(lastjoinrow==-1){
	joinrow=rn;
      }else{
	joinrow=lastjoinrow;
      };
      //Ignore rest of line;
      for(;*t&&*t!='\n'&&*t!=';';t++);
      continue;
    };
    for(;*t==' '||*t=='\t'||*t=='\r';t++);
    int *bm=new int[num];
    if(*t&&*t!='\n'&&*t!=';'){//Pattern present, delete previous pattern.
      row.clear();
      for(int i=0;i<num;i++){
	*(bm+i)=-1;
      };
      cur.clear();
    };
    int rf=-1;
    int rs=-1;
    int ps=0;
    for(;*t&&*t!='\n'&&*t!=';';){//read pattern
      if(isdigit(*t)){
	ps=atoi(t);
	if(ps<0){
	  ps=0;
	};
	if(ps>=num){
	  ps=num-1;//Not good.
	};
	for(;isdigit(*t);t++);
	for(;*t==' '||*t=='\t'||*t=='\r';t++);
	//	lastjoinrow=joinrow;
	if(rf!=-1){//Fill in from previous pattern
	  if(rf<ps){
	    for(int i=rf+1;i<=ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		cur.insert(1,&i);
	      };
	    };
	  }else{
	    for(int i=rf-1;i>=ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		cur.insert(1,&i);
	      };
	    };
	  };
	}else if(rs!=-1){
	  row.insert(cur);
	  cur.clear();
	  if(rs<ps){
	    for(int i=rs+1;i<ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		row.insert(1,&i);
	      };
	    };
	  }else{
	    for(int i=rs-1;i>ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		row.insert(1,&i);
	      };
	    };
	  };
	  cur.insert(1,&ps);
	}else{
	  cur.insert(1,&ps);
	  *(bm+ps)=1;
	};
	rs=-1;
	rf=-1;
      }else if(*t=='.'){
	rs=ps;
	t++;
      }else if(*t=='-'){
	rf=ps;
	t++;
      }else if(*t==','){
	row.insert(cur);
	cur.clear();
	t++;
      }else/* if(*t==':')*/{
	t++;
      };
    };
    if(cur.sz>0&&rf==-1){
      row.insert(cur);
      cur.clear();
    };
    if(rf!=-1){//Ends with a -
      ps=num-1;
      for(int i=rf+1;i<=ps;i++){	      
	if(*(bm+i)==-1){
	  *(bm+i)=1;
	  cur.insert(1,&i);
	};
      };
      row.insert(cur);
      cur.clear();
    }else if(rs!=-1){
      ps=num-1;
      for(int i=rs+1;i<=ps;i++){	      
	if(*(bm+i)==-1){
	  *(bm+i)=1;
	  row.insert(1,&i);
	};
      };
    };
    if(from!=rn){
      if(joinrow==-1){
	for(linkedlist<int>* rw=&row;rw!=NULL;rw=rw->next){
	  for(unsigned int i=0;i<rw->sz;i++){
	    *(cl+rn*(num+1)+1+*(rw->l+i))=nextcl;
	  };
	  if(rw->sz>0){
	    nextcl++;
	  };
	};
	*(cl+rn*(num+1))=1;
      }else{
	for(linkedlist<int>* rw=&row;rw!=NULL;rw=rw->next){
	  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	  for(unsigned int i=0;i<rw->sz;i++){
	    *(cl+rn*(num+1)+1+*(rw->l+i))=cno;
	  };
	};
	*(cl+rn*(num+1))=1;
	joinrow=-1;
      };
    };
    delete[] bm;
  };
  if(from!=-1){
    rn=params::numpars;
    if(from<rn){
      for(int i=from+1;i<rn;i++){
	if(*(cl+i*(num+1))==-1){
	  *(cl+i*(num+1))=1;
	  if(joinrow==-1){//New blocks for each row
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
	      };
	      if(rw->sz>0){
		nextcl++;
	      };
	    };
	  }else{//Use same blocks for all rows
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=cno;
	      };
	    };
	  };
	};
      };
    }else{
      for(int i=from-1;i>rn;i--){
	if(*(cl+i*(num+1))==-1){
	  *(cl+i*(num+1))=1;
	  if(joinrow==-1){//New blocks for each row
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
	      };
	      if(rw->sz>0){
		nextcl++;
	      };
	    };
	  }else{//Use same blocks for all rows
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=cno;
	      };
	    };
	  };
	};
      };
    };    
  };
  npars=nextcl;
  int pos=0;
  for(int i=0;i<params::numpars;i++){
    for(int j=1;j<=num;j++){
      if(*(cl+i*(num+1)+j)>=0){
	pos++;
      };
    };
  };
  pars=new int[pos+1];
  int ps=0;
  for(int parno=0;parno<npars;parno++){
    int sn=1;
    for(int i=0;i<params::numpars;i++){
      for(int j=1;j<=num;j++){
	if(*(cl+i*(num+1)+j)==parno){
	  *(pars+ps++)=sn*((j-1)*params::numpars+i);
	  sn=-1;
	};
      };
    };
  };
  *(pars+ps)=0;
  delete[] cl;
  cur.remove();
  row.remove();
};

void Mixture::assign(istream &s){
  int rn=0;
  int *cl=new int[(num+1)*params::numpars];
  for(int i=0;i<(num+1)*params::numpars;i++){
    *(cl+i)=-1;
  };
  int nextcl=0;
  int joinrow=-1;
  int lastjoinrow=-1;
  int from=-1;
  linkedlist<int> cur;
  linkedlist<int> row;
  char x[2];
  x[1]='\0';
  for(;s.good()&&!s.eof();){
    for(;s.good()&&!s.eof()&&(s.peek()==' '||s.peek()=='\t'||s.peek()=='\r'||s.peek()=='\n'||s.peek()==';');s.get());
    if(s.good()&&!s.eof()&&isdigit(s.peek())){
      rn=0;
      for(;s.good()&&!s.eof()&&isdigit(s.peek());){
	rn*=10;
	x[0]=s.get();
	rn+=atoi(x);
      };
      if(rn>=params::numpars){
	if(from==-1){
	  for(;s.good()&&!s.eof()&&s.peek()!='\n'&&s.peek()!=';';s.get());
	  continue;
	}else{
	  rn=params::numpars;
	};
	//Could give surprising behaviour. Should improve this.
      };
      for(;s.good()&&!s.eof()&&(s.peek()==' '||s.peek()=='\t'||s.peek()=='\r');s.get());
      lastjoinrow=joinrow;
      if(from!=-1){//Fill in from previous pattern
	if(from<rn){
	  for(int i=from+1;i<rn;i++){
	    if(*(cl+i*(num+1))==-1){
	      *(cl+i*(num+1))=1;
	      if(joinrow==-1){//New blocks for each row
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
		  };
		  if(rw->sz>0){
		    nextcl++;
		  };
		};
	      }else{//Use same blocks for all rows
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=cno;
		  };
		};
	      };
	    };
	  };
	}else{
	  for(int i=from-1;i>rn;i--){
	    if(*(cl+i*(num+1))==-1){
	      *(cl+i*(num+1))=1;
	      if(joinrow==-1){//New blocks for each row
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
		  };
		  if(rw->sz>0){
		    nextcl++;
		  };
		};
	      }else{//Use same blocks for all rows
		for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
		  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
		  for(unsigned int j=0;j<rw->sz;j++){
		    *(cl+i*(num+1)+1+*(rw->l+j))=cno;
		  };
		};
	      };
	    };
	  };
	};
      };
      from=-1;
    }else if(s.good()&&!s.eof()&&s.peek()=='.'){
      from=rn;
      s.get();
      //      for(;s.good()&&!s.eof()&&s.peek()!='\n'&&s.peek()!=';';s.get());
      //      continue;
    }else if(s.good()&&!s.eof()&&s.peek()=='|'){
      if(lastjoinrow==-1){
	joinrow=rn;
      }else{
	joinrow=lastjoinrow;
      };
      //Ignore rest of line;
      for(;s.good()&&!s.eof()&&s.good()&&!s.eof()&&s.peek()!='\n'&&s.peek()!=';';s.get());
      continue;
    };
    for(;s.good()&&!s.eof()&&(s.peek()==' '||s.peek()=='\t'||s.peek()=='\r');s.get());
    int *bm=new int[num];
    if(s.good()&&!s.eof()&&s.peek()!='\n'&&s.peek()!=';'){//Pattern present, delete previous pattern.
      row.clear();
      for(int i=0;i<num;i++){
	*(bm+i)=-1;
      };
      cur.clear();
    };
    int rf=-1;
    int rs=-1;
    int ps=0;
    for(;s.good()&&!s.eof()&&s.peek()!='\n'&&s.peek()!=';';){//read pattern
      if(isdigit(s.peek())){
	ps=0;
	for(;s.good()&&!s.eof()&&isdigit(s.peek());){
	  ps*=10;
	  x[0]=s.get();
	  ps+=atoi(x);
	};
	if(ps<0){
	  ps=0;
	};
	if(ps>=num){
	  ps=num-1;//Not good.
	};
	for(;s.good()&&!s.eof()&&isdigit(s.peek());s.get());
	for(;s.good()&&!s.eof()&&(s.peek()==' '||s.peek()=='\t'||s.peek()=='\r');s.get());
	//	lastjoinrow=joinrow;
	if(rf!=-1){//Fill in from previous pattern
	  if(rf<ps){
	    for(int i=rf+1;i<=ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		cur.insert(1,&i);
	      };
	    };
	  }else{
	    for(int i=rf-1;i>=ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		cur.insert(1,&i);
	      };
	    };
	  };
	}else if(rs!=-1){
	  row.insert(cur);
	  cur.clear();
	  if(rs<ps){
	    for(int i=rs+1;i<ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		row.insert(1,&i);
	      };
	    };
	  }else{
	    for(int i=rs-1;i>ps;i++){	      
	      if(*(bm+i)==-1){
		*(bm+i)=1;
		row.insert(1,&i);
	      };
	    };
	  };
	  cur.insert(1,&ps);
	}else{
	  cur.insert(1,&ps);
	  *(bm+ps)=1;
	};
	rs=-1;
	rf=-1;
      }else if(s.good()&&!s.eof()&&s.peek()=='.'){
	rs=ps;
	s.get();
      }else if(s.good()&&!s.eof()&&s.peek()=='-'){
	rf=ps;
	s.get();
      }else if(s.good()&&!s.eof()&&s.peek()==','){
	row.insert(cur);
	cur.clear();
	s.get();
      }else if(s.good()&&!s.eof()/*&&s.peek()==':'*/){
	s.get();
      };
    };
    if(cur.sz>0&&rf==-1){
      row.insert(cur);
      cur.clear();
    };
    if(rf!=-1){//Ends with a -
      ps=num-1;
      for(int i=rf+1;i<=ps;i++){	      
	if(*(bm+i)==-1){
	  *(bm+i)=1;
	  cur.insert(1,&i);
	};
      };
      row.insert(cur);
      cur.clear();
    }else if(rs!=-1){
      ps=num-1;
      for(int i=rs+1;i<=ps;i++){	      
	if(*(bm+i)==-1){
	  *(bm+i)=1;
	  row.insert(1,&i);
	};
      };
    };
    if(from!=rn){
      if(joinrow==-1){
	for(linkedlist<int>* rw=&row;rw!=NULL;rw=rw->next){
	  for(unsigned int i=0;i<rw->sz;i++){
	    *(cl+rn*(num+1)+1+*(rw->l+i))=nextcl;
	  };
	  if(rw->sz>0){
	    nextcl++;
	  };
	};
	*(cl+rn*(num+1))=1;
      }else{
	for(linkedlist<int>* rw=&row;rw!=NULL;rw=rw->next){
	  int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	  for(unsigned int i=0;i<rw->sz;i++){
	    *(cl+rn*(num+1)+1+*(rw->l+i))=cno;
	  };
	};
	*(cl+rn*(num+1))=1;
	joinrow=-1;
      };
    };
    delete[] bm;
  };
  if(from!=-1){
    rn=params::numpars;
    if(from<rn){
      for(int i=from+1;i<rn;i++){
	if(*(cl+i*(num+1))==-1){
	  *(cl+i*(num+1))=1;
	  if(joinrow==-1){//New blocks for each row
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
	      };
	      if(rw->sz>0){
		nextcl++;
	      };
	    };
	  }else{//Use same blocks for all rows
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=cno;
	      };
	    };
	  };
	};
      };
    }else{
      for(int i=from-1;i>rn;i--){
	if(*(cl+i*(num+1))==-1){
	  *(cl+i*(num+1))=1;
	  if(joinrow==-1){//New blocks for each row
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=nextcl;
	      };
	      if(rw->sz>0){
		nextcl++;
	      };
	    };
	  }else{//Use same blocks for all rows
	    for(linkedlist<int> *rw=&row;rw!=NULL;rw=rw->next){
	      int cno=*(cl+joinrow*(num+1)+1+*(rw->l));
	      for(unsigned int j=0;j<rw->sz;j++){
		*(cl+i*(num+1)+1+*(rw->l+j))=cno;
	      };
	    };
	  };
	};
      };
    };    
  };
  npars=nextcl;
  int pos=0;
  for(int i=0;i<params::numpars;i++){
    for(int j=1;j<=num;j++){
      if(*(cl+i*(num+1)+j)>=0){
	pos++;
      };
    };
  };
  pars=new int[pos+1];
  int ps=0;
  for(int parno=0;parno<npars;parno++){
    int sn=1;
    for(int i=0;i<params::numpars;i++){
      for(int j=1;j<=num;j++){
	if(*(cl+i*(num+1)+j)==parno){
	  *(pars+ps++)=sn*((j-1)*params::numpars+i);
	  sn=-1;
	};
      };
    };
  };
  *(pars+ps)=0;
  delete[] cl;
  cur.remove();
  row.remove();
};

void inline dummyreadmodel(istream& in){ 
  char x='-';
  in.get();
  int ql=0;
  char lq='\0';
  for(int level=1;in.good()&&level>0;x=in.get()){
    switch(x){
    case '}':
      if(ql==0){
	level--;
      };
      break;
    case '{':
      if(ql==0){
	level++;
      };
      break;
    case '\"':
      if(lq=='"'){
	ql--;
	lq=(ql>0)?'\'':'\0';
      }else{
	ql++;
	lq='\"';
      };
      break;
    case '\'':
      if(lq=='\''){
	ql--;
	lq=(ql>0)?'\"':'\0';
      }else{
	ql++;
	lq='\'';
      };
      break;
    case '\\':
      in.get();//Ignore next character
    };
  };
};

#include "ModelFields.h"

void readmodel(istream& in,int narg,const char * const * vnames,const char *const *vvals){
  int fct=1;
  for(int i=0;i<narg;i++){
    int fc=(strlen(*(vvals+i)))/strlen(*(vnames+i));
    if(fc>fct){
      fct=fc;
    };
  };
  for(;in.good()&&!in.eof()&&in.peek()!='}';){//for each line
    for(;in.good()&&!in.eof()&&(in.peek()==' '||in.peek()=='\t'||in.peek()=='\n'||in.peek()=='\r');in.get());
    char *field=readstring(in,"=",'\\');
    for(char *a=field+strlen(field)-1;*a==' '||*a=='\t'||*a=='\n'||*a=='\r';a--){
      *a='\0';
    };
    char *val=readstring(in,"\n",'\\');
    for(int ii=0;strlen(fieldnamesandtags[2*ii]);ii++){
      if(!strcmp(fieldnamesandtags[2*ii],field)){
	//Modify val by filling in parameters, stripping leading
	//whitespace and escaping as appropriate.
	char *modval=new char[strlen(val)*fct+1];
	if(*val=='#'){
	//Also allow basic numerical expressions. Numerical
	//expressions should be written in reverse polish notation.

	  //	  int ans=0;

	}else{
	  char *mv=val;
	  for(;*mv==' '||*mv=='\t'||*mv=='\n'||*mv=='\r';mv++);
	  for(char *a=val+strlen(val)-1;*a==' '||*a=='\t'||*a=='\n'||*a=='\r';a--){
	    *a='\0';
	  };
	  //Stripped leading and trailing spaces.
	  char lq='\0';
	  char *b=modval;
	  int *mtch=new int[narg];
	  for(int i=0;i<narg;i++){
	    *(mtch+i)=0;
	  };
	  for(char *a=mv;*a;a++){
	    if(lq=='\0'){
	      if(*a=='\"'||*a=='\''){
		lq=*a;
		for(int i=0;i<narg;i++){
		  *(mtch+i)=0;
		};
	      }else if(*a==' '||*a=='\t'||*a=='\r'){
		for(int i=0;i<narg;i++){
		  *(mtch+i)=0;
		};
	      }else{
		int cm=0;
		for(int i=0;i<narg;i++){
		  if(*(mtch+i)>=0){
		    if(*a==*(*(vnames+i)+*(mtch+i))){
		      (*(mtch+i))++;
		      if(*(mtch+i)==(signed int)strlen(*(vnames+i))){//match
			if(*(a+1)==' '||*(a+1)=='\t'||*(a+1)=='\r'||*(a+1)=='\0'||*(a+1)=='\"'||*(a+1)=='\''){
			  for(const char *aa=*(vvals+i);*aa;aa++){
			    *(b++)=*aa;
			  };
			  for(int i=0;i<narg;i++){
			    *(mtch+i)=0;
			  };
			  cm=1;
			}else{
			  *(mtch+i)=-1;
			  if(cm==0){
			    cm=-*(mtch+i);
			  };
			};
		      }else{
			cm=*(mtch+i);
		      };
		    }else{
		      if(cm==0){
			cm=-*(mtch+i);
		      };
		      *(mtch+i)=-1;
		    };
		  };
		};
		if(cm<=0){//no matching
		  for(char *aa=a+cm;aa<=a;aa++){
		    *(b++)=*aa;
		  };
		};
	      };
	    }else{
	      if(*a=='\\'){
		a++;
		*b++=*a;
	      }else{
		if(*a==lq){
		  lq='\0';  
		}else{
		  *b++=*a;
		};
	      };
	    };
	  };
	  *b='\0';
	  delete[] mtch;
	  if(lq!='\0'){
	    fatalError("Badly formated model file.");
	  };
	};
	variable* rec=vars.seektag(fieldnamesandtags[2*ii+1]);
	if(rec==NULL){
	  vars.addvar(fieldnamesandtags[2*ii],fieldnamesandtags[2*ii+1],new char*[1],setstr,modval,prntstr,delstr);
	}else{
	  rec->setval(modval);
	};
	delete[] modval;
	break;
      };
    };
    delete[] field;
    delete[] val;
  };
};

void passquotes(istream& in,char q){
  int ql=1;
  char lq=q;
  for(char a=in.get();in.good()&&!in.eof()&&ql>0;a=in.get()){
    if(a=='\''){
      if(a==lq){
	ql--;
	lq='\"';
      }else{
	ql++;
	lq='\'';
      };
    }else if(a=='"'){
      if(a==lq){
	ql--;
	lq='\'';
      }else{
	ql++;
	lq='\"';
      };
    }else if(a=='\\'){
      a=in.get();
    };
  };
};

void skipto(istream& in,char x){
  char a='0';
  for(;in.good()&&!in.eof()&&in.peek()!=x;a=in.get()){
    if(a=='\''||a=='\"'){
      passquotes(in,a);
    }else if(a=='\\'){
      a=in.get();
    };
  };
};


void thismodel(istream& in,const char *name,int narg,const char *const *vvals){
  for(;in.good()&&!in.eof()&&(in.peek()==' '||in.peek()=='\t'||in.peek()=='\n'||in.peek()=='\r');in.get());
  char x=in.get();
  if(x=='('){//Parameters
    char *pnames=readstring(in,")",'\\');	  
    int esc=0;
    int pno=1;
    for(char *pos=pnames;*pos;pos++){
      if(esc==0&&*pos==','){
	pno++;
      };
      esc=0;
      if(*pos=='\\'){
	esc=1;
      };
    };
    if(narg<pno){//too few parameters
      char *msg=new char[80+strlen(name)];
      strcpy(msg,"Too few parameters for model: \"");
      strcat(msg,name);
      strcat(msg,"\".");
      recoverableError(msg);
      delete[] msg;
    };
    char **vn=new char*[pno];
    int cpno=0;
    char *lpos=pnames;
    char *pos=pnames;
    for(;*pos;pos++){
      if(esc==0&&*pos==','){
	*(vn+cpno)=new char[pos-lpos+1];
	*pos='\0';
	strcpy(*(vn+cpno),lpos);
	lpos=pos+1;
	cpno++;
      };
      esc=0;
      if(*pos=='\\'){
	esc=1;
      };
    };
    *(vn+cpno)=new char[pos-lpos+1];
    *pos='\0';
    strcpy(*(vn+cpno),lpos);
    lpos=pos+1;
    cpno++;
    for(x=in.get();in.good()&&!in.eof()&&x!='{';x=in.get());
    readmodel(in,pno,vn,vvals);
    for(int i=0;i<pno;i++){
      delete[] *(vn+i);
    };
    delete[] vn;
    delete[] pnames;
  }else if(x=='{'){
    readmodel(in,0,NULL,vvals);  
  }else{//Badly formatted model file
    warning("Badly formatted model file: expected '{' or '('");
    skipto(in,'{');
    readmodel(in,0,NULL,vvals);  
  };
};

void readmodel(istream& in,const char *name,int narg,const char *const *vvals){
  for(;in.good()&&!in.eof();){
    for(;in.good()&&!in.eof()&&(in.peek()==' '||in.peek()=='\t'||in.peek()=='\n'||in.peek()=='\r');in.get());//Skip whitespace
    char x=in.get();
    if(x=='\"'){//Read until closing quote
      char *mname=readstring(in,"\"",'\\');
      if(!strcmp(mname,name)){//This model
	thismodel(in,name,narg,vvals);
	delete[] mname;
	return;
      }else{//ignore this model
	skipto(in,'{');
	dummyreadmodel(in);
      };
      delete[] mname;
    }else{//First character of model name
      if(x=='\\'){//start of escape sequence
	x=in.get();
	if(x==*name){
	  char *mname=readstring(in,"( \t\r\n{",'\\');
	  if(!strcmp(mname,name+1)){//This model
	    thismodel(in,name,narg,vvals);
	    delete[] mname;
	    delete[] mname;
	    return;
	  }else{
	    skipto(in,'{');
	    dummyreadmodel(in);
	    delete[] mname;
	  };
	}else{
	  skipto(in,'{');
	  dummyreadmodel(in);
	};
      }else if(x==*name){//possible match
	char *mname=readstring(in,"( \t\r\n",'\\');
	if(!strcmp(mname,name+1)){//This model
	  thismodel(in,name,narg,vvals);
	  delete[] mname;
	  return;
	}else{//ignore this model
	  skipto(in,'{');
	  dummyreadmodel(in);
	delete[] mname;
	};
      }else{
	skipto(in,'{');
	dummyreadmodel(in);
      };
    };
  };
  char *msg=new char[80+strlen(name)];
  strcpy(msg,"Model: \"");
  strcat(msg,name);
  strcat(msg,"\" not defined.");
  recoverableError(msg);
  delete[] msg;
};
