/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <math.h>
#include <time.h>
using namespace std;

extern ostream& comm;//commentry;

#include "ErrorHandler.h"
#include "Macros.h"
#include "Miscellaneous.h"
#include "SignalHandler.h"
#include "CommandLine.h"
#include "Matrix.h"
#include "Sequence.h"
#include "Parameters.h"
#include "Likelihood.h"
#include "Optimisation.h"
#include "Tree.h"
#include "TreeLikelihood.h"

// Code needed for optimisation.

class Mixture{
  //determines which parameters are the same across all models 
  //and which parameters are constant.
  int num;
  int npars;
  int *pars;// Negative values indicate a continuation of the previous
	    // parameter. List is terminated by 0.
public:
  Mixture(){num=0;npars=0;pars=NULL;};
  Mixture(int n){num=n;npars=0;pars=NULL;};
  Mixture(const char *s,int n);
  Mixture(istream& s,int n);
  ~Mixture(){if(pars!=NULL){delete[] pars;};};
  int valid(){return pars!=NULL;};
  void setnum(int n){num=n;};
  void assign(const char *s);
  void assign(istream &s);
  friend class Mixtreelikelihood;
  void setpars(const long double *vals,long double *in);
  void sumpars(const long double *in,long double *out);
  void sumparsdecstep(const long double *in,long double *out,int firststep);
  void sumparstrisquare(const long double *dt,long double *to,int dim,int offset=0);
  void print();
};

void Mixture::print(){
  for(int i=0;i<params::numpars;i++){
    for(int j=0;j<num;j++){
      int pr=-1;
      int pn=-1;
      for(int *p=pars;*p;p++){
	if(*p>0){
	  pn++;
	};
	if(*p==j*params::numpars+i){
	  pr=pn;
	}else if(*p==-j*params::numpars-i){
	  pr=pn;
	};
      };
      cout<<pr<<"\t";
    };
    cout<<"\n";
  };
};

void Mixture::setpars(const long double *vals,long double *in){
  //Fills the values vals into the appropriate places in the list in.
  int *pos=pars;
  for(int i=0;i<npars;i++){
    *(in+*(pos++))=*(vals+i);
    for(;*(pos)<0;pos++){
      *(in-*(pos))=*(vals+i);
    };
  };
};

void Mixture::sumpars(const long double *in,long double *out){
  int *pos=pars;
  for(int i=0;i<npars;i++){
    *(out+i)=*(in+*(pos++));
    for(;*(pos)<0;pos++){
      *(out+i)+=*(in-*(pos));
    };
  };
};

int decsum(int a,int b){
  return (b*(b+1)-(b-a)*(b-a+1))/2;
};

void Mixture::sumparsdecstep(const long double *in,long double *out,int firststep){
  int *pos=pars;
  for(int i=0;i<npars;i++){
    *(out+i)=*(in+decsum(*(pos++),firststep));
    for(;*(pos)<0;pos++){
      *(out+i)+=*(in+decsum(-*(pos),firststep));
    };
  };
};

void Mixture::sumparstrisquare(const long double *dt,long double *to,int dim,int offset){
  int *pos=pars;
  for(int i=0;i<npars;i++){
    int *ps=pars;
    for(int j=0;j<npars;j++){
      //      cout<<*(pos)<<":\n\t"<<*(ps)<<"\t";
      *(to+i*(dim)+j)=*(dt+tri(*pos,*(ps++),(num*params::numpars+offset)));
      for(;*(ps)<0;ps++){
	//	cout<<*(ps)<<"\t";
	*(to+i*(dim)+j)+=*(dt+tri(*pos,-*ps,(num*params::numpars+offset)));
      };
    };
    for(pos++;*(pos)<0;pos++){
      //      cout<<"\n"<<*(pos);
      ps=pars;
      for(int j=0;j<npars;j++){
	*(to+i*(dim)+j)+=*(dt+tri(-*pos,*(ps++),(num*params::numpars+offset)));
	for(;*(ps)<0;ps++){
	  *(to+i*(dim)+j)+=*(dt+tri(-*pos,-*ps,(num*params::numpars+offset)));
	};
      };
    };
    //    cout<<"\n\n";
  };
};


class Mixtreelikelihood : public treelk{//Should extend this to functionplus at some point.
  tree T;
  int num;
  params *p;
  long double *prob;
  int varprobs;
  sequence *d;
  Factmatrix *f;
  long double *lastval;
  long double *vals;
  int *variablepars;//Makes a list of parameters allowed to vary.
  int dir;
  Mixture m;//This will replace variablepars.
  int testlast(const long double *x);
  Mixtreelikelihood(const Mixtreelikelihood& T);//Disable copying
  inline void calculatevals(const long double *x);
public:
  Mixtreelikelihood(int mx,const char *tr,const Realmatrix& Q,ifstream& data);
  Mixtreelikelihood(int mx,const char *tr,ifstream& data);
  void setpars(const Realmatrix& Q);
  void setpars(ifstream& I);
  ~Mixtreelikelihood(){delete[] d;delete[] p;delete[] prob;T.up=NULL;T.remove();for(int i=0;i<num;i++){(f+i)->remove();};delete[] f;delete[] lastval;if(vals!=NULL){delete[] vals;};delete[] variablepars;};
  void readpars(istream &in){for(int i=0;i<num;i++){(p+i)->readpars(in);*(f+i)=(p+i)->fmatrix();};};
  void getx(long double *out);
  long double evaluate(const long double *x);
  long double *sitelikes(){return NULL/*T.Mixsitelikelihoods(f,d,p)*/;};
  int numsites(){return d->length;};
  void deriv(const long double *x,long double *out);
  void hessian(const long double *x,long double *out);
  /*  long double directionalDeriv(const long double *x,const long double *v);
  long double directionalSecDer(const long double *x,const long double *v);
  void approxHessian(const long double *x,long double *out);
  int LineSearch(long double *x,long double lastval,const long double *dir,int numpos);*/
  long double quickEvaluate(const long double *x);
  int edges(){return T.edges();};
  int posvals(){return T.edges()+varprobs;};
  void constpars(){dim=T.edges();*variablepars=-1;};
  void printtree(ostream& out=cout){T.print(out);};
  void printdata(){T.printdata(cout,d);};
  void normalise();//{T.scale(p->normalise());};//Need to also modify other parameters.
  void normalisejustpars();//{for(int i=0;i<num;i++){(p+i)->normalise();};};//Need to also modify other parameters.
  void orthpars(){params::orthmats();};
  void printall(ostream& out=cout,int *valpars=NULL);
  void parsimony();
  void empirical(int type=0){long double *epi=emppi(d,T.leaves(),type);for(int i=0;i<num;i++){(p+i)->setpi(epi);};/*cout<<"Empirical frequencies:\n";for(int j=0;j<NumCodons;j++){cout<<*(epi+j)<<"\t";};cout<<"\n";*/ delete[] epi;};
  void setprob(long double *pr){for(int i=0;i<num-1;i++){*(prob+i)=*(pr+i);};};
  void setmix(const char*s){m.assign(s);dim=T.numed+num-1+m.npars;};
  void setmix(ifstream& s){m.assign(s);dim=T.numed+num-1+m.npars;};
  void setmixfile(const char* s,const char *const*pt=NULL);
  void setvarprobs(int i){dim+=varprobs-i;varprobs=i;};
  void setfixedpars(long double *vls);
  void tstat(long double *x,ostream& out);
  int* tstatsel(long double *x,long double threshold=TSTAT5);
  void selectpars(int *sel);
  void undoselect(int *sel,int lastpars);
  void printtreedev(const Realmatrix &oi,ostream& out=cout){long double *std=new long double[T.numed]; for(int i=0;i<T.numed;i++){*(std+i)=sqrt(*(oi.entries+i*(oi.sz+1)));};T.print(std,out);delete[] std;};
  long double *truepars(int *valpars=NULL);
  RealmatrixT oi(long double *x);
};

long double *Mixtreelikelihood::truepars(int *valpars){
  return NULL;
  //NEED TO WRITE THIS.
};

RealmatrixT Mixtreelikelihood::oi(long double *x){
  //NEED TO WRITE THIS.
  return RealmatrixT(0);
};

void Mixtreelikelihood::undoselect(int *sel,int lastpars){
  //NEED TO WRITE THIS.
};

void Mixtreelikelihood::selectpars(int *sel){
  //NEED TO WRITE THIS.
};


int* Mixtreelikelihood::tstatsel(long double *x,long double threshold){
  return NULL;//NEED TO WRITE THIS.
};

void Mixtreelikelihood::tstat(long double *x,ostream& out){
  Realmatrix hmat(T.numed+params::numpars*num+num);
};

void Mixtreelikelihood::normalisejustpars(){
  long double *pp=new long double[num];
  long double tp=0;
  for(int i=0;i<num-1;i++){
    tp+=*(prob+i);
  };
  *(pp+num-1)=1/(tp+1);
  for(int i=0;i<num-1;i++){
    *(pp+i)=(*(prob+i))*(*(pp+num-1));
  };
  if(params::numpars>0){
    long double ans=0;
    for(int i=0;i<num;i++){
      long double l=0;
      for(int j=0;j<NumCodons;j++){
	for(int k=0;k<NumCodons;k++){
	  if(j!=k){
	    l+=*((p+i)->pi+j)*(*((p+i)->pi+k))*(*((p+i)->R.entries+j*NumCodons+k));
	  };
	};
      };
      ans+=l*(*(pp+i));
    };
    delete[] pp;
    long double la=log(ans)/(*(p->XX->entries+1));
    for(int i=0;i<num;i++){
      *(p+i)->coeff-=la;
    (p+i)->setR();
    };
  };
};

void Mixtreelikelihood::normalise(){
  long double *pp=new long double[num];
  long double tp=0;
  for(int i=0;i<num-1;i++){
    tp+=*(prob+i);
  };
  *(pp+num-1)=1/(tp+1);
  for(int i=0;i<num-1;i++){
    *(pp+i)=(*(prob+i))*(*(pp+num-1));
  };
  if(params::numpars>0){
    long double ans=0;
    for(int i=0;i<num;i++){
      long double l=0;
      for(int j=0;j<NumCodons;j++){
	for(int k=0;k<NumCodons;k++){
	  if(j!=k){
	    l+=*((p+i)->pi+j)*(*((p+i)->pi+k))*(*((p+i)->R.entries+j*NumCodons+k));
	  };
	};
      };
      ans+=l*(*(pp+i));
  };
    delete[] pp;
    long double la=log(ans)/(*(p->XX->entries+1));
    for(int i=0;i<num;i++){
      *(p+i)->coeff-=la;
      (p+i)->setR();
    };
    T.scale(ans);
  };
};

void Mixtreelikelihood::setfixedpars(long double *vls){
  int *fx=new int[num*params::numpars];
  for(int i=0;i<num*params::numpars;i++){
    *(fx+i)=1;
  };  
  for(int *mpr=m.pars;*mpr;mpr++){
    int j=abs(*(mpr));
    *(fx+j)=0;
  };
  long double *y=new long double[params::numpars];
  long double *z=new long double[params::numpars];
  for(int i=0;i<num;i++){
    for(int j=0;j<params::numpars;j++){
      *(z+j)=*((p+i)->coeff+j);
    };
    ((Realmatrix)params::A.inverse()).act(z,y);
    for(int j=0;j<params::numpars;j++){
      if(*(fx+i*params::numpars+j)){
	*(y+j)=*(vls++);
      };
    };
    params::A.act(y,z);
    (p+i)->setCoeff(z);
  };
  delete[] fx;
  delete[] y;    
  delete[] z;    
};

void Mixtreelikelihood::getx(long double *out){
  if(m.valid()){
    int *pos=m.pars;
    for(int i=0;i<m.npars;i++){
      int pv=*(pos++);
      *(out+i+T.numed+varprobs)=(p+pv/params::numpars)->coeff[pv%params::numpars];
      for(;*pos<0;pos++);
    };
  }else{
    for(int i=0;i<dim-T.edges()-varprobs;i++){
      *(out+i+T.numed+varprobs)=(p+*(variablepars+i)/params::numpars)->coeff[*(variablepars+i)%params::numpars];
    };
  };
  for(int i=0;i<varprobs;i++){
    *(out+T.numed+i)=*(prob+i);
  };
  T.putedgelengths(out);
};

int search(const char *const*path,const char *filename,ifstream& in);

void Mixtreelikelihood::setmixfile(const char* s,const char *const*pt){
  ifstream in;
  while(!search(pt,s,in)){
    char *msg=new char[strlen(s)+45];
    strcpy(msg,"Unable to open mixture file: \"");
    strcat(msg,s);
    strcat(msg,"\"");
    recoverableError(msg,msgcode(0,3),&filemenu);
    delete[] msg;
  };
  m.assign(in);
  in.close();
  dim=T.numed+varprobs+m.npars;
};

Mixtreelikelihood::Mixtreelikelihood(int mx,const char *tr,const Realmatrix& Q,ifstream& data){
  if(mx<2){
    throw mx;
  };
  p=new params[mx];
  T=tree(tr);
  T.filledges();
  num=mx;
  varprobs=num-1;
  m.setnum(num);
  prob=new long double[num-1];
  f=new Factmatrix[mx];
  p->getpars(Q);
  for(int i=1;i<num;i++){
    *(p+i)=*p;
  };
  for(int i=0;i<num;i++){
    *(f+i)=(p+i)->fmatrix();
  };
  d=getdata(data,T);
  dim=T.edges()+num*p->numpars+num-2;
  if(p->numpars==0){
    dim++;
  };
  lastval=new long double[dim];
  vals=NULL;variablepars=new int[num*params::numpars+1];
  for(int i=0;i<num*params::numpars-1;i++){
    *(variablepars+i)=i+1;
  };
  *(variablepars+params::numpars-1)=-1;
  dir=0;
};


Mixtreelikelihood::Mixtreelikelihood(int mx,const char *tr,ifstream& data){
  if(mx<2){
    throw mx;
  };
  p=new params[mx]; //Need to create params before tree to ensure numpars is known.
  T=tree(tr);
  T.filledges();
  num=mx;
  varprobs=num-1;
  m.setnum(num);
  prob=new long double[num-1];
  f=new Factmatrix[mx];
  d=getdata(data,T);
  dim=T.edges()+num*p->numpars+num-2;
  if(p->numpars==0){
    dim++;
  };
  lastval=new long double[dim];
  vals=NULL;
  variablepars=NULL;
  dir=0;
};

void Mixtreelikelihood::setpars(const Realmatrix& Q){
  if(p->numpars==0){
    dim=T.numed+varprobs;
  };
  p->getpars(Q);
  for(int i=1;i<num;i++){
    *(p+i)=*p;
  };
  for(int i=0;i<num;i++){
    *(f+i)=(p+i)->fmatrix();
  };
  vals=NULL;
  if(variablepars!=NULL){
    delete[] variablepars;
  };
  variablepars=new int[num*params::numpars+1];
  for(int i=0;i<num*params::numpars-1;i++){
    *(variablepars+i)=i+1;
  };
  if(params::numpars>0){
    *(variablepars+num*params::numpars-1)=-1;
  };
};

void Mixtreelikelihood::setpars(ifstream& I){
  if(p->numpars==0){
    dim=T.numed+varprobs;
  };
  p->readpars(I);
  for(int i=1;i<num;i++){
    *(p+i)=*p;
  };
  for(int i=0;i<num;i++){
    *(f+i)=(p+i)->fmatrix();
  };
  vals=NULL;
  if(variablepars!=NULL){
    delete[] variablepars;
  };
  variablepars=new int[num*params::numpars+1];
  for(int i=0;i<num*params::numpars-1;i++){
    *(variablepars+i)=i+1;
  };
  if(params::numpars>0){
    *(variablepars+num*params::numpars-1)=-1;
  };
};

void Mixtreelikelihood::parsimony(){
  //Uses parsimony to seed the branch lengths.
  Realmatrix R=p->Rmatrix();
  for(int i=0;i<NumCodons;i++){
    for(int j=0;j<NumCodons;j++){
      if(i!=j){
	*(R.entries+i*NumCodons+j)*=sqrt(*(p->pi+i)*(*(p->pi+j)));
      };
    };
  };
  for(int i=0;i<NumCodons;i++){
    *(R.entries+i*(NumCodons+1))=0;
    for(int j=0;j<NumCodons;j++){
      if(i!=j){
	*(R.entries+i*(NumCodons+1))-=*(R.entries+i*NumCodons+j);
      };
    };
  };
  Realmatrix C=R.getresistance();
  T.parsimonybl(C,d);
};

void Mixtreelikelihood::printall(ostream &out,int *valpars){
  //prints all coefficients, even fixed ones.
  T.print(out);
  out<<"\n\n";
  long double tp=0;
  for(int i=0;i<num-1;i++){
    tp+=*(prob+i)+1;
  };
  tp=1/(tp+1);
  if(params::numpars>0){
    for(int i=0;i<num-1;i++){
      out<<"Site class "<<i<<" probability="<<(*(prob+i)+1)*tp<<":\n\n";
      out<<"Parameter";
      out.width(25);
      out<<"Coefficient";
      out.width(40);
      out<<"Exponential of Coefficient\n\n";
      long double *tc=(p+i)->truepars();
      for(int j=0;j<params::numpars;j++){
	out.width(5);
	out<<j;
	out.width(29);
	out<<*(tc+j);
	out.width(30);
	out<<exp(*(tc+j))<<"\n";
      };
      out<<"\n\n";
      delete[] tc;
    };
    out<<"Site class "<<num-1<<" probability="<<tp<<":\n\n";
    out<<"Parameter";
    out.width(25);
    out<<"Coefficient";
    out.width(40);
    out<<"Exponential of Coefficient\n\n";
    long double *tc=(p+num-1)->truepars();
    for(int j=0;j<params::numpars;j++){
      out.width(5);
      out<<j;
      out.width(29);
      out<<*(tc+j);
      out.width(30);
      out<<exp(*(tc+j))<<"\n";
    };
    delete[] tc;
    out<<"\n\n";
  };
  //  delete[] x;
};

long double Mixtreelikelihood::quickEvaluate(const long double *x){
  if(dim>T.numed){
    long double *y=new long double[num*params::numpars];
    if(m.valid()){
      for(int i=0;i<num;i++){
	for(int j=0;j<params::numpars;j++){
	  *(y+i*params::numpars+j)=*((p+i)->coeff+j);
	};
      };
      m.setpars(x+T.numed+varprobs,y);
    }else{
      int j=0;
      for(int i=0;i<dim-varprobs-T.numed;i++){
	for(int k=(i==0)?0:(*(variablepars+i-1)+1);k<*(variablepars+i);k++){
	  *(y+k)=(p+j)->coeff[k%num];
	};
	*(y+*(variablepars+i))=*(x+T.numed+varprobs+i);
      };
    }
    for(int i=0;i<num;i++){
      (p+i)->setCoeff(y+i*params::numpars);
      *(f+i)=(p+i)->fmatrix();
    };
    delete[] y;
  };
  T.setedgelengths(x);
  for(int i=0;i<varprobs;i++){
    *(prob+i)=*(x+T.numed+i);
  };
  return T.Mixlikelihood(f,d,num,prob,p);
};


int Mixtreelikelihood::testlast(const long double *x){
  if(lastval==NULL){
    return 0;
  };
  for(int i=0;i<dim;i++){
    if(NONZEROSTSQ(*(x+i)-*(lastval+i))){
      return 0;
    };
  };
  return 1;
};

inline void Mixtreelikelihood::calculatevals(const long double *x){
  if(vals!=NULL){
    delete[] vals;
  };
  //Need to change to new values x
  if(dim>T.numed){
    long double *y=new long double[num*params::numpars];
    if(m.valid()){
      for(int i=0;i<num;i++){
	for(int j=0;j<params::numpars;j++){
	  *(y+i*params::numpars+j)=*((p+i)->coeff+j);
	};
      };
      m.setpars(x+T.numed+varprobs,y);
    }else{
      int j=0;
      for(int i=0;i<dim-varprobs-T.numed;i++){
	for(int k=(i==0)?0:(*(variablepars+i-1)+1);k<*(variablepars+i);k++){
	*(y+k)=(p+j)->coeff[k%num];
	};
	*(y+*(variablepars+i))=*(x+T.numed+varprobs+i);
      };
    };
    for(int i=0;i<num;i++){
      (p+i)->setCoeff(y+i*params::numpars);
      *(f+i)=(p+i)->fmatrix();
    };
    delete[] y;    
  };
  for(int i=0;i<varprobs;i++){
    *(prob+i)=*(x+T.numed+i);
  };
  for(int i=0;i<dim;i++){
    *(lastval+i)=*(x+i);
  };
  T.setedgelengths(x);
  vals=Mixcalchess(T,f,d,num,prob,p);
  //    vals=T.hessian(f,d,p);
};

long double Mixtreelikelihood::evaluate(const long double *x){
  if(vals==NULL||!this->testlast(x)){//Need to recalculate
    this->calculatevals(x);
  };
  return *vals;
};


void Mixtreelikelihood::deriv(const long double *x,long double *out){
  if(vals==NULL||!this->testlast(x)){//Need to recalculate
    this->calculatevals(x);
  };
  for(int i=0;i<T.edges();i++){
    *(out+i)=*(vals+i+1);
  };
  if(m.valid()){
    m.sumpars(vals+T.numed+1,out+T.numed+varprobs);
  }else{
    for(int i=0;i<(dim-varprobs-T.edges());i++){
      *(out+T.numed+varprobs+i)=*(vals+T.numed+*(variablepars+i)+1);
    };
  };
  for(int i=0;i<varprobs;i++){
    *(out+T.numed+i)=*(vals+1+T.numed+num*params::numpars+i);
  };
};

void Mixtreelikelihood::hessian(const long double *x,long double *out){
  if(vals==NULL||!this->testlast(x)){//Need to recalculate
    this->calculatevals(x);
  };
  int acc=T.numed+num*params::numpars+num;
  for(int i=0;i<T.numed;i++){
    for(int j=i;j<T.numed;j++){
      *(out+i*dim+j)=*(vals+acc);
      *(out+j*dim+i)=*(vals+acc);
      acc++;
    };
    for(int j=0;j<varprobs;j++){
      *(out+i*dim+T.numed+j)=*(vals+acc+num*params::numpars);
      *(out+i+(T.numed+j)*dim)=*(vals+acc+num*params::numpars);
      acc++;
    };
    if(m.valid()){
      m.sumpars(vals+acc-varprobs,out+i*dim+T.numed+varprobs);
      for(int j=0;j<dim-varprobs-T.numed;j++){
	*(out+(j+T.numed+varprobs)*dim+i)=*(out+i*dim+T.numed+j+varprobs);
      };
    }else{
      for(int j=0;j<dim-varprobs-T.numed;j++){
	*(out+i*dim+T.numed+j+varprobs)=*(vals+acc+num-1-varprobs+*(variablepars+j));
	*(out+(j+T.numed+varprobs)*dim+i)=*(vals+acc+num-1-varprobs+*(variablepars+j));
      };
    };
    acc+=num*params::numpars+num-1-varprobs;
  };
  acc=T.numed*((T.numed+1)/2+num*params::numpars+num)+num*params::numpars+num;
  if(m.valid()){
    m.sumparstrisquare(vals+acc,out+(T.numed+varprobs)*dim+(T.numed+varprobs),dim,num-1);
    acc+=num*params::numpars;
    for(int j=0;j<varprobs;j++){
      m.sumparsdecstep(vals+acc,out+(T.numed+j)*dim+T.numed+varprobs,params::numpars*num+num-2);
      for(int i=T.numed+varprobs;i<dim;i++){
	*(out+i*dim+j+T.numed)=*(out+(j+T.numed)*dim+i);
      };
      acc++;
    };
  }else{
    for(int i=0;i<dim-varprobs-T.numed;i++){
      for(int k=(i>0)?*(variablepars+i-1):0;k<*(variablepars+i);k++){
	acc+=num*params::numpars+varprobs-k;//*(variablepars+i);
      };
      for(int j=i;j<dim-varprobs-T.numed;j++){
	*(out+(i+T.numed)*dim+j+T.numed)=*(vals+acc+*(variablepars+j)-*(variablepars+i));
	*(out+(j+T.numed)*dim+i+T.numed)=*(vals+acc+*(variablepars+j)-*(variablepars+i));
      };
      for(int j=0;j<varprobs;j++){
	*(out+(i+T.numed)*dim+dim+j-varprobs)=*(vals+acc+num*params::numpars-*(variablepars+i)+j);
	*(out+(dim+j-varprobs)*dim+i+T.numed)=*(vals+acc+num*params::numpars-*(variablepars+i)+j);
      };
    };
  };
  acc=(T.numed+num*params::numpars+num)*(T.numed+num*params::numpars+num+1)/2-num*(num-1)/2;
  for(int i=0;i<varprobs;i++){
    for(int j=i;j<varprobs;j++){
      *(out+(T.numed+i)*dim+T.numed+j)=*(vals+acc);
      *(out+(T.numed+j)*dim+T.numed+i)=*(vals+acc);
      acc++;
    };
    acc+=num-1-varprobs;
  };
};



