/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
using namespace std;

#include "Macros.h"


/* Classes and routines for dealing with DNA sequences */

int codon(char *seq){
  //Anything except ACGT is treated as unknown. Maybe better in future
  //to do something cleverer.
  //  cout<<"Converting sequence \""<<*seq<<*(seq+1)<<*(seq+2)<<"\"\n";
  int a=*seq=='A'?8:*seq=='C'?4:*seq=='G'?12:*seq=='T'?0:128;
  a+=*(seq+1)=='A'?32:*(seq+1)=='C'?16:*(seq+1)=='G'?48:*(seq+1)=='T'?0:256;
  a+=*(seq+2)=='A'?2:*(seq+2)=='C'?1:*(seq+2)=='G'?3:*(seq+2)=='T'?0:64;
  if(a>33){//Stop codons TAA=34 and TAG=35
    a-=2;
  };
  if(a>47){//Stop codon TGA=50
    a--;
  };
  return a;
};

char *displayCodon(int cod){
  char *a=new char[3];
  if(cod>47){
    cod++;
  };
  if(cod>33){
    cod+=2;
  };
  if(cod>255){
    *(a+1)='-';
    cod-=256;
    if(cod>127){
      *a='-';
      *(a+2)=cod>=130?(cod==131?'G':'A'):(cod==129?'C':'T');
    }else if(cod>63){
      *(a+2)='-';
      *a=cod>=71?(cod==75?'G':'A'):(cod==67?'C':'T');
    }else{
      *a=cod>=8?(cod>=12?'G':'A'):(cod>=4?'C':'T');
      cod%=4;
      *(a+2)=cod>=2?(cod==3?'G':'A'):(cod==1?'C':'T');
    };
  }else if(cod>127){
    *a='-';
    cod-=127;
    if(cod>63){
      *(a+2)='-';
      cod-=64;
      *(a+1)=cod>=32?(cod>=48?'G':'A'):(cod>=16?'C':'T');
    };
  }else if(cod>63){
    *(a+2)='-';
    cod-=64;
    *(a+1)=cod>=32?(cod>=48?'G':'A'):(cod>=16?'C':'T');
    cod%=16;
    *a=cod>=8?(cod>=12?'G':'A'):(cod>=4?'C':'T');
    cod%=4;
  }else{
    *(a+1)=cod>=32?(cod>=48?'G':'A'):(cod>=16?'C':'T');
    cod%=16;
    *a=cod>=8?(cod>=12?'G':'A'):(cod>=4?'C':'T');
    cod%=4;
    *(a+2)=cod>=2?(cod==3?'G':'A'):(cod==1?'C':'T');
  };
  return a;
};

class sequence{
public:
  int length; //Maybe make these data private
  char *seq;  //But for the moment, this is fine.
  sequence(){length=0;seq=NULL;};
  sequence(char *s){int l=strlen(s);length=l/CodonLength;seq=new char[l];for(int i=0;i<l;i++){*(seq+i)=*(s+i);};};
  ~sequence(){if(seq!=NULL){delete[] seq;};};
  void assign(char *s){int l=strlen(s);length=l/CodonLength;seq=new char[l];for(int i=0;i<l;i++){*(seq+i)=*(s+i);};};
  void display() const;
  void copypos(int pos,char *target) const {for(int i=0;i<CodonLength;i++){*(target+i)=*(seq+pos*CodonLength+i);};};
  int getLength() const {return length;};
  void assign(ifstream &in);
  void assign(ifstream &in,int len);
  friend ostream& operator <<(ostream& out,const sequence& s);
};

ostream& operator <<(ostream& out,const sequence& s){
  for(int i=0;i<s.length;i++){
    for(int j=0;j<CodonLength;j++){
      out<<*(s.seq+i*CodonLength+j);
    };
    out<<" ";
  };
  return out;
};

char *comparepos(sequence *s,int pos,int numseq){
  char *a=new char[numseq*CodonLength];
  for(int i=0;i<numseq;i++){
    (s+i)->copypos(pos,a+i*CodonLength);
  };
  return a;
};

void sequence::display() const {
  for(int i=0;i<length*CodonLength;i++){
    cout<<*(seq+i);
  };
};

long double *emppi(sequence *data,int len,int type=0){
  long double *ans=new long double[NumCodons];
  long double nucfreq[12];
  int total=0;
  for(int i=0;i<12;i++){
    nucfreq[i]=0;
  };
  for(int i=0;i<NumCodons;i++){
    *(ans+i)=0;
  };
  for(int i=0;i<len;i++){
    total+=(data+i)->length;
    for(int j=0;j<(data+i)->length;j++){      
      int num=codon((data+i)->seq+j*CodonLength);
      if(num<NumCodons){
	*(ans+num)+=1;
      }else if(num<125){
	num-=61;
	num/=4;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=(i/4==num)?(1/4):0;
	};
      }else if(num<189){
	num-=125;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=((i%4==num%4)&&(i/16==num/16))?1/4:0;
	};
      }else if(num<253){
	num-=189;
	num/=16;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=(i/16==num)?1/16:0;
	};
      }else if(num<317){
	num-=253;      
	num%=16;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=(i%16==num)?1/4:0;
	};
      }else if(num<381){
	num-=317;      
	num%=16;
	num/=4;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=((i%16)/4==num)?1/16:0;
	};
      }else if(num<445){
	num-=381;      
	num%=4;
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=(i%4==num)?1/16:0;
	};
      }else{
	for(int i=0;i<NumCodons;i++){
	  *(ans+i)+=1/NumCodons;
	};
      };
      for(int k=0;k<CodonLength;k++){
	switch(*((data+i)->seq+j*CodonLength+k)){
	case 'A':
	  nucfreq[k*4+2]+=1;
	  break;
	case 'C':
	  nucfreq[k*4+1]+=1;
	  break;
	case 'G':
	  nucfreq[k*4+3]+=1;
	  break;
	case 'T':
	  nucfreq[k*4]+=1;
	  break;
	default:
	  nucfreq[k*4]+=0.25;
	  nucfreq[k*4+1]+=0.25;
	  nucfreq[k*4+2]+=0.25;
	  nucfreq[k*4+3]+=0.25;
	};
      };
    };
  };
  switch(type){
  case 0:  //Empirical Codon Frequencies
    for(int i=0;i<NumCodons;i++){
      *(ans+i)/=total;
      if(*(ans+i)==0){
	*(ans+i)=1e-16;
      };
    };
    break;
  case 1: { //F3x4
    for(int i=0;i<12;i++){
      nucfreq[i]/=total;
    };
    long double tfreq=0;
    for(int i=0;i<NumCodons;i++){
      int j=(i>33)?i+2:i;
      j=(j>49)?j+1:j;
      *(ans+i)=nucfreq[(j%16)/4]*nucfreq[4+j/16]*nucfreq[8+j%4];
      tfreq+=*(ans+i);
    };
    for(int i=0;i<NumCodons;i++){
      *(ans+i)/=tfreq;
    };
  };
    break;
  case 2:  {//F1x4
    for(int i=0;i<4;i++){
      nucfreq[i]+=nucfreq[i+4]+nucfreq[i+8];
      nucfreq[i]/=total*3;
    };
    long double tfreq=0;
    for(int i=0;i<NumCodons;i++){
      int j=(i>33)?i+2:i;
      j=(j>49)?j+1:j;
      *(ans+i)=nucfreq[(j%16)/4]*nucfreq[j/16]*nucfreq[j%4];
      tfreq+=*(ans+i);
    };
    for(int i=0;i<NumCodons;i++){
      *(ans+i)/=tfreq;
    };
  };
    break;
  case 3:  //Fequal
    for(int i=0;i<NumCodons;i++){
      *(ans+i)=(long double)1/61;
    };
  };
  return ans;
};

