/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <float.h>
#include <stdio.h>
#include <fstream>
#include <iostream>
#include <signal.h>
#include <stdarg.h>
#include <string.h>
using namespace std;
#define BACKTRACESIZE 20

class flushlock{
  //For something like the state, threads notify the lock while
  //accessing data. When one thread wants to save, all other threads
  //are stopped. The saving thread must first wait until all other
  //threads have finished with the relevant data.
public:
  flushlock();
  void access();
  void access_finished();
  void flush();
  void flush_finished();
};

extern flushlock flushstate;


class state{
  //Used for storing the important variables representing the
  //computations already performed. This way, the output can be
  //analysed to improve debugging. Also, the program can later be set
  //to resume from where it was up to.
  char *filename;
  //  int numdata; this is just strlen type.
  const char *retid; //identification string to indicate where program was up to
  const void **data;
  char *type;
  long unsigned int *numitems;//Usually, this is a long unsigned int, telling how
		 //many items pointed to by data.
  int *no;
  int next;
public:
  state(const char* pos){filename=NULL;data=NULL;type=NULL;numitems=NULL;retid=pos;no=NULL;next=0;};
  void setfile(const char* name){if(filename!=NULL){delete[] filename;};filename=new char[strlen(name)+1];strcpy(filename,name);};
  void save();//used on termination to allow resumption
  int addvar(char ty,long unsigned int numitems,const void *add,const char *pos="Unspecified");//adds a new variable to the state.
  void remvar(int num,const char *pos=NULL,int nv=1);//adds a new variable to the state.
  void resizedata(int n,unsigned long int sz,const char *pos);//extends the amount of
      //data of interest.
  void clear(int k);//just keep one variable					
  void clear();//tidy up when information is processed.
  ~state(){this->clear();if(filename!=NULL){delete[] filename;};};
};

class recoverState{
  //Used to recover a saved state, and resume program execution.
  ifstream read;
  int pos;
  char *type; //indicates the types of the data stored in the file.
public:
  char *retid; //identification string to indicate where program was up to
  int num(char t){int ans=0;for(char *p=type+pos;*p;p++){if(*p==t){ans++;};};return ans;};
  recoverState(char *filename);
  void load(int n,...);//used for resumption after process interruption.
  void load(int n,long unsigned int *max,void **address);//used for resumption after process interruption.
  void remove(){delete[] retid;delete[] type;read.close();};//For tidying up after finishing recovery
};

void endrecover();

extern state* CurrentState; //This stores the current state of the program,
		     //so that if the program is interrupted, the
		     //signal handler can save the state
		     //appropriately.

void setstate(state *st);

extern recoverState* Recover;//If recovery is required, this will
				//store the necessary recovery
				//variable.

inline void endrecover(){Recover=NULL;};

inline void recover(char *filename){Recover=new recoverState(filename);if(Recover->retid==NULL){delete Recover;Recover=NULL;};};

void TerminatingSignalHandler(int sig,int sigtype);

int SetUpSignalHandlers(const void *variables=NULL);
