/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "Matrix.h"
#include "Tree.h"
#include "Sequence.h"
#include "Parameters.h"
#include "CommandLine.h"
#include "Miscellaneous.h"
#include <math.h>
#include <fstream>
#include <iostream>
using namespace std;


class simulationtree{//like tree, but just used for simulation, so no lists for likelihood.
public:
  int n;
  int numed;
  const char *label;
  Realmatrix *P;
  simulationtree *subtree;
  long unsigned int *down;//basic downwards lists.
//public:
  simulationtree(){n=0;numed=0;P=NULL;subtree=NULL;down=NULL;};
  simulationtree(const tree &t,const params &p,const Factmatrix &f);
  void del();
  ~simulationtree(){this->del();};
  void assign(const tree &t,const params &p,const Factmatrix &f);
  void simulate(char **s,long unsigned int n,const params &p);
  void simsite(int start,char**s,long unsigned int offset,int pos=0);
  int leaves(){if(n==0){return 1;};int l=0;for(int i=0;i<n;i++){l+=(subtree+i)->leaves();};return l;};
  void printsequences(ostream &out,const char *const* s,long unsigned int len,long unsigned int offset,int &pos);
};

void simulationtree::del(){
  //  for(int i=0;i<n;i++){
  //    (subtree+i)->del();
  //  };
  //  if(n>0){
    delete[] subtree;
    delete[] P;
    delete[] down;
    //  };
};

simulationtree::simulationtree(const tree &t,const params &p,const Factmatrix &f){
  n=t.n;
  numed=t.numed;
  label=t.label;
  P=new Realmatrix[n];
  subtree=new simulationtree[n];
  down=new long unsigned int[NumCodons];
  long double *ex=new long double[NumCodons];
  for(int i=0;i<n;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+j)=exp(*(f.D.entries+j)*(*(t.length+i)));
    };
    (subtree+i)->assign(*(t.subtree+i),p,f);
    (P+i)->sz=NumCodons;
    (P+i)->entries=new long double[NumCodonsSq];
    for(int j=0;j<NumCodons;j++){
      for(int k=0;k<NumCodons;k++){
	*((P+i)->entries+j*NumCodons+k)=0;
	for(int l=0;l<NumCodons;l++){
	  *((P+i)->entries+j*NumCodons+k)+=*(f.gamma.entries+j*NumCodons+l)*(*(ex+l))*(*(f.gammainv.entries+l*NumCodons+k));
	};
      };
    };  
  };
  delete[] ex;
};

void simulationtree::assign(const tree &t,const params &p,const Factmatrix &f){
  n=t.n;
  numed=t.numed;
  label=t.label;
  P=new Realmatrix[n];
  subtree=new simulationtree[n];
  down=new long unsigned int[NumCodons];
  long double *ex=new long double[NumCodons];
  for(int i=0;i<n;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+j)=exp(*(f.D.entries+j)*(*(t.length+i)));
    };
    (subtree+i)->assign(*(t.subtree+i),p,f);
    (P+i)->sz=NumCodons;
    (P+i)->entries=new long double[NumCodonsSq];
    for(int j=0;j<NumCodons;j++){
      for(int k=0;k<NumCodons;k++){
	*((P+i)->entries+j*NumCodons+k)=0;
	for(int l=0;l<NumCodons;l++){
	  *((P+i)->entries+j*NumCodons+k)+=*(f.gamma.entries+j*NumCodons+l)*(*(ex+l))*(*(f.gammainv.entries+l*NumCodons+k));
	};
      };
    };  
  };
  delete[] ex;
};

void simulationtree::simulate(char **s,long unsigned int n,const params&p){
  for(long unsigned int i=0;i<n/3;i++){
    long double r=((long double) rand())/RAND_MAX;
    long double rsum=*p.pi;
    int j=0;
    for(;rsum<r;j++){
      rsum+=*(p.pi+j+1);
    };
    this->simsite(j,s,i,0);
  };
};

void simulationtree::simsite(int start,char**s,long unsigned int offset,int pos){
  if(n==0){
    char *dc=displayCodon(start);
    *(*(s+pos)+3*offset)=*dc;
    *(*(s+pos)+3*offset+1)=*(dc+1);
    *(*(s+pos)+3*offset+2)=*(dc+2);
    delete[] dc;
    return;
  };
  int ps=pos;
  for(int i=0;i<n;i++){
    long double l=((long double) rand())/RAND_MAX;
    long double rsum=*((P+i)->entries+start*NumCodons);
    int j=0;
    for(;rsum<l;j++){
      rsum+=*((P+i)->entries+start*NumCodons+j+1);
    };
    (subtree+i)->simsite(j,s,offset,ps);
    ps+=(subtree+i)->leaves();
  };    
};

void simulationtree::printsequences(ostream &out,const char *const* s,long unsigned int len,long unsigned int offset,int &pos){
  if(n==0){
    out<<label<<"   ";
    for(unsigned long int i=0;i<len;i++){
      out<<*(*(s+pos)+offset+i);
    };
    pos++;
    out<<"\n";
    return;
  };
  for(int i=0;i<n;i++){
    (subtree+i)->printsequences(out,s,len,offset,pos);
  };
};

void simulate(const char* basefilename,long unsigned int numfiles, long unsigned int seqlength,const tree &t,params &p){
  Factmatrix f=p.fmatrix();
  simulationtree s(t,p,f);
  int l=t.leaves();
  char **sq=new char*[l];
  for(int i=0;i<l;i++){
    *(sq+i)=new char[numfiles*seqlength];
  };
  s.simulate(sq,numfiles*seqlength,p);
  ofstream out(basefilename);
  char *filename=new char[strlen(basefilename)+20];//Easily enough
  //Put metadata in basefile
  out<<t;//.print(out);
  out<<"\n\n";
  if(params::numpars>0){
    out<<"Parameter";
    out.width(25);
    int *sel=NULL;
    variable *rec=vars.seektag("-parameterselection");
    if(rec!=NULL){
      sel=selection(*(char **)rec->value);
      sortlist(sel);
      out<<"Number in file";
      out.width(25);
    };
    out<<"Coefficient\n";
    long double *tc=p.truepars();
    for(int j=0;j<params::numpars;j++){
      out.width(5);
      out<<j;
      out.width(29);
      if(sel!=NULL){
	out<<((j==0)?0:(*(sel+j-1)));
	out.width(25);
      };
      out<<*(tc+j)<<"\n";
    };
    delete[] tc;
    out<<"\n\n";
  };
  out.close();
  for(long unsigned int flno=0;flno<numfiles;flno++){
    sprintf(filename,"%s%06lu",basefilename,flno);
    out.open(filename);
    out<<l<<"\t"<<seqlength<<"\n";
    int p=0;
    s.printsequences(out,sq,seqlength,flno*seqlength,p);
    out.close();
  };
  f.remove();
  for(int i=0;i<l;i++){
    delete[] *(sq+i);
  };
  delete[] sq;
  delete[] filename;
};
