/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef TREE_H
#define TREE_H 1

#include <iostream>
using namespace std;
#include "Parameters.h"
#include "Sequence.h"

class tree{
public:
  int n;
  int numed;
  long double *length;
  tree *subtree;
  char *label;
  long double *down;//basic downwards lists.
  long double *down2;//downwards lists from top of branch
  long double *down3;//transformed downwards lists
  long double *up;//basic upwards lists
  long double *up2;//transformed upwards lists
  long double *up3;//upwards lists from bottom of branch
  long double *extra;
  long double *extra2;
  long double *extra3;//Just for parameters
//public:
  tree(){n=0;length=NULL;subtree=NULL;label=NULL;down=NULL;down2=NULL;down3=NULL;up=NULL;up2=NULL;up3=NULL;extra=NULL;extra2=NULL;extra3=NULL;};
  tree(const tree& T){n=T.n;length=T.length;subtree=T.subtree;label=T.label;up=T.up;up2=T.up2;up3=T.up3;down=T.down;down2=T.down2;down3=T.down3;extra=T.extra;extra2=T.extra2;extra3=T.extra3;};
  void operator=(tree T){n=T.n;length=T.length;subtree=T.subtree;label=T.label;up=T.up;up2=T.up2;up3=T.up3;down=T.down;down2=T.down2;down3=T.down3;extra=T.extra;extra2=T.extra2;extra3=T.extra3;};
  tree(int l,const char *s);
  tree(const char *s);
  void remove();
  long double *putedgelengths(long double *x){long double *ret=x;for(int i=0;i<n;i++){*ret=*(length+i);ret=(subtree+i)->putedgelengths(ret+1);};return ret;};
  const long double *setedgelengths(const long double *x){const long double *ret=x;for(int i=0;i<n;i++){*(length+i)=*ret;ret=(subtree+i)->setedgelengths(ret+1);};return ret;};
  void filledges(){numed=n;for(int i=0;i<n;i++){(subtree+i)->filledges();numed+=(subtree+i)->numed;};};
  int edges() const {int l=n;for(int i=0;i<n;i++){l+=(subtree+i)->edges();};return l;};
  int leaves() const {if(n==0){return 1;}else{int l=0;for(int i=0;i<n;i++){l+=(subtree+i)->leaves();};return l;};};
  void print(ostream& out=cout) const;
  void print(long double *bl,ostream& out=cout) const;
  friend ostream &operator <<(ostream &o,const tree &T);
  int findlabel(char *lab) const;
  void liklistdown(const Factmatrix& f,const sequence *data,int i,const long double *ex,long double len=-1); 
                              //builds the list of likelihoods
  void liklistdownfast(const Factmatrix& f,const sequence *data,int i,const long double *ex,const Realmatrix& gis,long double len=-1); 
  void liklistup(const params &p,const Factmatrix& f,const long double *ex,long double *above=NULL,long double len=0);
                              //builds the lists of upwards likelihoods.
  void printlists(int level,int start,int u,int st=0,int ed=0) const;
  void printdata(ostream& out,sequence* d) const;
  void getlengths(long double*& out) const{for(int i=0;i<n;i++){*(out++)=*(length+i);(subtree+i)->getlengths(out);};};
  long double *getlengths() const{long double *ans=new long double[this->edges()];long double *a2=ans;this->getlengths(a2);return ans;};
  const tree *findedge(int e,long double &len) const;
  long double *hessian(const Factmatrix& f,const sequence *data,const params& p);
  long double *approxHessian(const Factmatrix& f,const sequence *data,const params& p);
  long double *derivs(const Factmatrix& f,const sequence *data,const params& p);
  long double likelihood(const Factmatrix& f,const sequence *data,const params& p);
  long double fastlikelihood(const Factmatrix& f,const sequence *data,const params& p);
  long double *sitelikelihoods(const Factmatrix& f,const sequence *data,const params& p);
  void shiftbranchlength(int no,long double eps){for(int i=0;i<n;i++){if(no--==0){*(length+i)+=eps;return;}else if(no<(subtree+i)->edges()){(subtree+i)->shiftbranchlength(no,eps);return;}else{no-=(subtree+i)->edges();};};};
  void printspecies(ostream& out,int* no) const;
  void setextra(int ed);
  //Routines for calculating directional derivatives.
  long double *directionalDerivs(const Factmatrix& f,const sequence *data,const params& p,const long double *dir);
  void debugneg(long double val,ostream& out=cout,int siteno=-1,int l=0);
  void scale(long double x){for(int i=0;i<n;i++){(subtree+i)->scale(x);*(length+i)*=x;};};
  void parsimonybl(const Realmatrix &C,sequence *data);
  void loglengths();

  //Mixture Models
  long double *Mixhessian_threaded(const Factmatrix* f,const sequence *data,int num,long double *prob,const params* p,int st);
  long double Mixlikelihood(const Factmatrix* f,const sequence *data,int num,const long double *prob,const params* p);
};

sequence *getdata(ifstream& f,const tree& T);
long double *calchess(tree &t, const Factmatrix& f,const sequence *data,const params& p);
void setnumthreads(int i);

long double *Mixcalchess(tree &t, const Factmatrix* f,const sequence *data,int num,const long double *prob,const params* p);
#endif
