/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef TREELIKELIHOOD_H
#define TREELIKELIHOOD_H

#include "Tree.h"
#include "Parameters.h"
#include "Sequence.h"
#include "Matrix.h"
#include "Optimisation.h"

// Code needed for optimisation.

class treelk : virtual public function{//Interface for using tree likelihoods for normal and mixture models in same way.
 public:
  virtual ~treelk(){};
  virtual void setpars(const Realmatrix& Q)=0;
  virtual void setpars(ifstream& I)=0;
  virtual void readpars(istream &in)=0;
  virtual void getx(long double *out)=0;
  virtual long double evaluate(const long double *x)=0;
  virtual long double *sitelikes()=0;
  virtual int numsites()=0;
  virtual void deriv(const long double *x,long double *out)=0;
  virtual void hessian(const long double *x,long double *out)=0;
  virtual long double quickEvaluate(const long double *x)=0;
  virtual int edges()=0;
  virtual int posvals()=0;
  virtual void constpars()=0;
  virtual void printtree(ostream& out=cout)=0;
  virtual void printdata()=0;
  virtual void normalise()=0;
  virtual void normalisejustpars()=0;
  virtual void orthpars()=0;
  virtual void printall(ostream& out=cout,int *valpars=NULL)=0;
  virtual void parsimony()=0;
  virtual void empirical(int type=0)=0;
  virtual void setfixedpars(long double *vls)=0;
  virtual void tstat(long double *x,ostream& out)=0;
  virtual int* tstatsel(long double *x,long double threshold=TSTAT5)=0;
  virtual void selectpars(int *sel)=0;
  virtual void undoselect(int *sel,int lastpars)=0;
  virtual void printtreedev(const Realmatrix &oi,ostream& out=cout)=0;
  virtual long double *truepars(int *valpars=NULL)=0;
  virtual RealmatrixT oi(long double *x)=0;
};



class logbllike;
class logplike;

class treelikelihood : public functionplus, public treelk{
  tree T;
  params p;
  sequence *d;
  Factmatrix f;
  long double *lastval;
  long double *vals;
  int *variablepars;//Makes a list of parameters allowed to vary.
  int dir;
  int testlast(const long double *x);
  treelikelihood(const treelikelihood& T);//Disable copying
public:
  treelikelihood(const char *tr,const Realmatrix& Q,ifstream& data){T=tree(tr);T.filledges();p.getpars(Q);f=p.fmatrix();d=getdata(data,T);dim=T.edges()+p.numpars-1;if(p.numpars==0){dim++;};lastval=new long double[dim];vals=NULL;variablepars=new int[params::numpars+1];for(int i=0;i<params::numpars-1;i++){*(variablepars+i)=i+1;};*(variablepars+params::numpars-1)=-1;dir=0;};
  treelikelihood(const char *tr,ifstream& data){T=tree(tr);T.filledges();d=getdata(data,T);dim=T.edges()+p.numpars-1;if(p.numpars==0){dim++;};lastval=new long double[dim];vals=NULL;variablepars=NULL;dir=0;};
  void setpars(const Realmatrix& Q){if(p.numpars==0){dim=T.numed;};p.getpars(Q);f=p.fmatrix();vals=NULL;if(variablepars!=NULL){delete[] variablepars;};variablepars=new int[params::numpars+1];for(int i=0;i<params::numpars-1;i++){*(variablepars+i)=i+1;};if(params::numpars>0){*(variablepars+params::numpars-1)=-1;};};
  void setpars(ifstream& I){if(p.numpars==0){dim=T.numed;};p.readpars(I);f=p.fmatrix();vals=NULL;if(variablepars!=NULL){delete[] variablepars;};variablepars=new int[params::numpars+1];for(int i=0;i<params::numpars-1;i++){*(variablepars+i)=i+1;};if(params::numpars>0){*(variablepars+params::numpars-1)=-1;};};
  ~treelikelihood(){delete[] d;T.up=NULL;T.remove();f.remove();delete[] lastval;if(vals!=NULL){delete[] vals;};delete[] variablepars;};
  void readpars(istream &in){p.readpars(in);f=p.fmatrix();};
  void getx(long double *out){for(int i=0;i<dim-T.edges();i++){*(out+i+T.edges())=p.coeff[*(variablepars+i)];};T.putedgelengths(out);};
  long double evaluate(const long double *x);
  long double *sitelikes(){return T.sitelikelihoods(f,d,p);};
  int numsites(){return d->length;};
  void deriv(const long double *x,long double *out);
  void hessian(const long double *x,long double *out);
  long double directionalDeriv(const long double *x,const long double *v);
  long double directionalSecDer(const long double *x,const long double *v);
  void approxHessian(const long double *x,long double *out);
  int LineSearch(long double *x,long double lastval,const long double *dir,int numpos);
  long double quickEvaluate(const long double *x);
  int edges(){return T.edges();};
  int posvals(){return T.edges();};
  void constpars(){dim=T.edges();*variablepars=-1;};
  void printtree(ostream& out=cout){T.print(out);};
  void printtreedev(const Realmatrix &oi,ostream& out=cout){long double *std=new long double[T.numed]; for(int i=0;i<T.numed;i++){*(std+i)=sqrt(*(oi.entries+i*(oi.sz+1)));};T.print(std,out);delete[] std;};
  void printdata(){T.printdata(cout,d);};
  void normalise(){if(params::numpars>0){T.scale(p.normalise());};};
  void normalisejustpars(){if(params::numpars>0){p.normalise();};};
  void orthpars(){p.orthmats();};
  void printall(ostream& out=cout,int *valpars=NULL);
  void parsimony();
  void empirical(int type=0){long double *epi=emppi(d,T.leaves(),type);p.setpi(epi);/*cout<<"Empirical frequencies:\n";for(int j=0;j<NumCodons;j++){cout<<*(epi+j)<<"\t";};cout<<"\n";*/ delete[] epi;};
  void setfixedpars(long double *vls);
  void tstat(long double *x,ostream& out);
  int* tstatsel(long double *x,long double threshold=TSTAT5);
  void selectpars(int *sel);
  void undoselect(int *sel,int lastnum);
  long double *truepars(int *valpars=NULL){return p.truepars();};
  RealmatrixT oi(long double *x);
  logbllike optbranchlength(int branch,int site1,int site2);
  logplike optpar(int par,int branch,int site1,int site2);
};

class logbllike : public onedfunction{
  int sites;
  long double *ev;
  long double *coeffs;
public:
  logbllike(int s=0){sites=s;ev=new long double[NumCodons];coeffs=new long double[NumCodons*sites];};
  void setsites(int s){sites=s;if(coeffs!=NULL){delete[] coeffs;};coeffs=new long double[NumCodons*sites];};
  void fillsite(int n,long double *v){for(int i=0;i<NumCodons;i++){*(coeffs+n*NumCodons+i)=*(v+i);};};//No safety checking!
  void setev(long double* e){for(int i=0;i<NumCodons;i++){*(ev+i)=*(e+i);};};
  long double evaluate(long double x);
  long double deriv(long double x);
  long double secder(long double x);
};

class logplike : public onedfunction{
  int sites;
  int pno;
  params *p;
  long double *coeffs;
  long double *y;
  long double len;
public:
  logplike(params *pa,int s=0){p=pa;sites=s;coeffs=new long double[NumCodons*sites*2];y=new long double[params::numpars];for(int i=0;i<params::numpars;i++){*(y+i)=p->coeff[i];};};
  void setsites(int s){sites=s;if(coeffs!=NULL){delete[] coeffs;};coeffs=new long double[NumCodons*sites*2];};
  void fillsite(int n,long double *v,long double *w){for(int i=0;i<NumCodons;i++){*(coeffs+n*NumCodons*2+i)=*(v+i);*(coeffs+(n*2+1)*NumCodons+i)=*(w+i);};};//No safety checking!
  void setpar(params *pa){p=pa;};
  long double evaluate(long double x);
  long double deriv(long double x);
  long double secder(long double x);
  friend logplike treelikelihood::optpar(int par,int branch,int site1,int site2);
};
#endif
